import numpy as np
import numpy.linalg as la

# 10**(-15) will have redundant transvec 
indep_vec_lim=10**(-32)
    
def makeOrthonormal(vecs,lim=indep_vec_lim):
    '''Decompose vectors 'vecs' into orthonormal set using Gram-Schmidt algorithm.
    Discard linearly dependent vectors. Works in real and fourier case.

    Yujie: It uses transpose!

    '''

    # print "Calculating Gram-Schmidt orthonormal basis of vectors..."     
    
    gram = []
 
    for v in vecs.T:
        gramproj = np.zeros_like(v,dtype=float)
        for g in gram:
            gramproj += g*np.vdot(g,v)
        v = v - gramproj            
        if la.norm(v) < np.sqrt(len(v))*lim:
            pass
            # print "\tDependent vector found, independent norm: %.1e." %(la.norm(v))
        else:
            # print "\tNew basis vector found, independent norm: %.1e." %(la.norm(v))               
            v = v/la.norm(v)      
            gram.append(v)              
    
    gram = np.asarray(gram).T
    # print "\tDimension of orthonormal basis: %d"%(np.shape(gram)[1])
    return gram   

def realMakeOrthonormal(vecs,lim=indep_vec_lim):
    '''
    Yujie: Same as makeOrthonormal, but no transpose
    '''

    # print "Calculating Gram-Schmidt orthonormal basis of vectors..."     
    
    gram = []
 
    for v in vecs:
        gramproj = np.zeros_like(v,dtype=float)
        for g in gram:
            gramproj += g*np.vdot(g,v)
        v = v - gramproj            
        if la.norm(v) < np.sqrt(len(v))*lim:
            pass
            # print "\tDependent vector found, independent norm: %.1e." %(la.norm(v))
        else:
            # print "\tNew basis vector found, independent norm: %.1e." %(la.norm(v))               
            v = v/la.norm(v)      
            gram.append(v)              
    
    gram = np.asarray(gram)
    # print "\tDimension of orthonormal basis: %d"%(np.shape(gram)[1])
    return gram   

# def makeOrthonormalNew(vecs,lim=indep_vec_lim): # Yujie, shuffling first vector to be taken, in order to get full basis
#     '''Decompose vectors 'vecs' into orthonormal set using Gram-Schmidt algorithm.
#     Discard linearly dependent vectors. Works in real and fourier case.'''

#     # print "Calculating Gram-Schmidt orthonormal basis of vectors..."     
#     gram = []

#     vecs2=np.copy(vecs)
#     for i in range(1,len(vecs2.T)):
#         print i, len(vecs2.T)
#         break2flag=True
#         for v in vecs2.T:
#             v=v/la.norm(v)
#             gramproj = np.zeros_like(v,dtype=float)
#             for g in gram:
#                 gramproj += g*np.vdot(g,v)
#             v = v - gramproj            
#             if la.norm(v) < np.sqrt(len(v))*lim:
#                 # tempvec=np.copy(vecs2.T[i])
#                 # vecs2.T[i]=np.copy(vecs2.T[0])
#                 # vecs2.T[0]=np.copy(tempvec)
#                 # print "shuffle ",i," to 0"
#                 vecs2=(np.random.permutation(vecs2.T)).T
#                 print 'shuffle'
#                 break2flag=False
#                 break
#             else:
#                 # print "\tNew basis vector found, independent norm: %.1e." %(la.norm(v))               
#                 v = v/la.norm(v)      
#                 gram.append(v)
#         if (break2flag):break
            
#     gram = np.asarray(gram).T
#         # print "\tDimension of orthonormal basis: %d"%(np.shape(gram)[1])
#     return gram   

    
def getTranslations2D(veclength,hasTSM):
    '''Make normalized pure translation vectors.
    Works in real and fourier case.'''
    ndim,dispindc,angindc = getDispInd2D(hasTSM)
    transentry = 1/np.sqrt(veclength/ndim)
    transs = []
    for ind in [d for d in dispindc if d!= -1]:
        trans = np.zeros(veclength)
        trans[ind::ndim]= transentry 
        transs.append(trans)
        
    if not transs: return None
    return np.asarray(transs).T   

def getDispInd2D(hasTSM=[True,False,False]): 
    ''' Finds degrees of freedom per point and their indices,
        given whether tension/shear/torsion exist.
        Works in real and fourier case.'''        
    xind=-1
    yind=-1
    angind=-1
    if hasTSM[1]:
        ndim = 3
        xind = 0
        yind = 1
        angind = 2
    elif hasTSM[0]:
        ndim = 2
        xind = 0
        yind = 1        
        if hasTSM[2]:
            ndim = 3
            angind = 2
    elif hasTSM[2]:
        ndim = 1
        angind = 0
    else: 
        ndim = 0
    return ndim,(xind,yind),angind


def removeTranslations2D(vecs,lim=indep_vec_lim):
    ''''For a set of displacement vectors (vecs) consisting of a subset of (x,y,angle) displacements
    (determined by hasTSM), construct uniform translation vectors. Orthonormalize the combined set (translations),(vecs).
    Return the orthonormalized (full!) set. Works in real and fourier case.'''
    # print "Separating uniform translations out of vector span 'vecs'..."

    transs = getTranslations2D(len(vecs),hasTSM=[True,False,False])   # [true,false,false] is apprppriate for spring networks without rotational DOF
    
    if transs is None: 
        print "\tOnly angle displacements in vector. Can't subtract translations."
        return vecs
    else:
        vecs = makeOrthonormal(vecs,lim)
        dim = np.shape(vecs)[1]
        for i,trans in enumerate(transs.T):
            vecs = makeOrthonormal(np.concatenate((trans[:,None],vecs),axis=1),lim)
            dimdv = np.shape(vecs)[1]
            # if dim == dimdv: print "\tUniform translation %d in span 'vecs'."%(i)
            # else: print "\tUniform translation %d not in span 'vecs'."%(i)
            
    return vecs

