import numpy as np
import scipy
from scipy import linalg as la
from scipy.sparse import lil_matrix,csc_matrix,csr_matrix,bmat
from scipy.sparse import linalg as las
from numpy.linalg import matrix_rank
from copy import deepcopy
from topomesh import vis2d

_default_smallest = 20
_close_to_zero = 1e-34

def zeropad(arr, padshape):
    '''pad numpy array with zeros along given axes'''
    newshape = tuple(np.array(arr.shape) + np.array(padshape))
    newarray = np.zeros(newshape,dtype=arr.dtype)
    newarray[:arr.shape[0], :arr.shape[1]] = arr
    return newarray

def null(A, eps=1e-12):
    u, s, vh = la.svd(A)
    null_mask = (s <= eps)
    null_space = scipy.compress(null_mask, vh, axis=0)
    return s[null_mask], scipy.transpose(null_space)

def rigiditymatrix(mesh,normed=False):
    r = []
    pts = mesh.points()
    bds = mesh.Bonds
    dim = len(pts[0])
    for bd in bds:
        row = np.zeros(dim*mesh.N)
        dp = mesh.dr(bd.p2,bd.p1)
        if normed: dp = dp/la.norm(dp)
        for i in range(dim):
            row[dim*bd.p1+i] = dp[i]
            row[dim*bd.p2+i] = -dp[i]
        r.append(row)
    return np.matrix(r)
    
def trussrigiditymatrix(mesh,torsion=True):
    '''
        a: cross-section area of truss
        i: second area moment of truss
        Compbook XIVA p 122
    '''
    r = []
    pts = mesh.points()
    bds = mesh.Bonds
    for bd in bds:
        dp = mesh.dr(bd.p1,bd.p2)   # NOTE sign here is opposite that of ordinary rigidity mx, for consistency with notes and literature
        L = la.norm(dp)
        
        dp = dp/L   # dp[0] is cos of rotation angle for element, dp[1] is sin
        
        # first row for this bond: axial stretching/compression
        row = np.zeros(3*mesh.N)
        row[3*bd.p1] = dp[0]
        row[3*bd.p1+1] = dp[1] 
        row[3*bd.p2] = -dp[0]
        row[3*bd.p2+1] = -dp[1]
        
        r.append(row)
        
        # second row: shear stretching/compression, subtracting out zero rotation
        row = np.zeros(3*mesh.N)
        
        row[3*bd.p1] = -dp[1]
        row[3*bd.p1+1] = dp[0]
        row[3*bd.p1+2] = L/2.
         
        row[3*bd.p2] = dp[1]
        row[3*bd.p2+1] = -dp[0]
        row[3*bd.p2+2] = L/2.
        
        r.append(row)
        
        if torsion:
            # third row: torsion at point 1
            row = np.zeros(3*mesh.N)
            row[3*bd.p1+2] = 1
            row[3*bd.p2+2] = -1
            r.append(row)
        
    return np.matrix(r)
    
def trusskmatrix(mesh,a=1,i=1,torsion=True):
    '''
        a: cross-section area of truss
        i: second area moment of truss
    '''
    blocks = []
    pts = mesh.points()
    bds = mesh.Bonds
    for bd in bds:
        r = []
        L = la.norm(mesh.dr(bd.p2,bd.p1))
        
        k = a/L     # spring constant for axial defs
        q = i/L**3  # moment for bending and shear defs
        
        if torsion:
            # each bond has 4 rows
            r = np.diag([k,12*q,L*L*q])
            
        else:
            # only two rows, two tensions (axial and transverse)
            r = np.diag([k,12*q])
        
        blocks.append(r)
        
    return np.matrix(la.block_diag(*blocks)).T

def rigiditymatrix_sparse(mesh,normed=False):
    r = []
    pts = mesh.points()
    dim = len(pts[0])
    bds = mesh.Bonds
    data=[]
    indices=[]
    indptr=[]
    for bd in bds:
        indptr.append(len(indices))
        dp = mesh.dr(bd.p2,bd.p1)
        if normed: dp = dp/la.norm(dp)
        for i in range(dim):
            indices.append(dim*bd.p1+i)
            data.append(dp[i])
            indices.append(dim*bd.p2+i)
            data.append(-dp[i])
        
    indptr.append(len(indices))
    return csr_matrix((np.array(data),indices,indptr),shape=(len(bds),len(pts)*dim))


def dynamicalmatrix(mesh):
    r = rigiditymatrix(mesh)
    return r.T*r                # matrix multiplication

def tensionmatrix(mesh):
    r = rigiditymatrix(mesh)
    return r*r.T                # matrix multiplication

def hamiltonian(mesh): # kane and lubensky hamiltonian
    r1 

def rigiditymodes(mesh,eps=1e-12):
    return null(rigiditymatrix(mesh),eps)

def modes(mx,smallest=None):
    eigval,eigvec = la.eig(mx)
    eigvec = np.array(eigvec).T
    
    if smallest is None:
        sortedargs = np.argsort(np.real(eigval))
        return np.real(eigval)[sortedargs], np.real(eigvec)[sortedargs]
        
    else:
        sortedargs = np.argsort(np.real(eigval)**2)
        return np.real(eigval)[sortedargs][:smallest], np.real(eigvec)[sortedargs][:smallest]

def modes_symm(mx,smallest=None):
    eigval,eigvec = la.eigh(mx)
    eigvec = np.array(eigvec).T
    
    if smallest is None:
        sortedargs = np.argsort(np.real(eigval))
        return np.real(eigval)[sortedargs], np.real(eigvec)[sortedargs]
        
    else:
        sortedargs = np.argsort(np.real(eigval)**2)
        return np.real(eigval)[sortedargs][:smallest], np.real(eigvec)[sortedargs][:smallest]  # Yujie: take real


def modes_sparse(mx,smallest=_default_smallest,which='LM',checkrank=False,**kwargs):
    eigval,eigvec = las.eigs(mx,smallest,sigma=_close_to_zero,which=which,**kwargs)
    eigvec = np.array(eigvec).T
    
    # check rank of imaginary part
    if checkrank and matrix_rank(np.imag(eigvec) != 0):
        # find conjugate pairs
        pairs = [(i,j) for i,ei in enumerate(eigval) for j,ej in enumerate(eigval) if (i < j and ei.conjugate() == ej)]
        print pairs
        for i,j in pairs:
            vi = (eigvec[i]+eigvec[j])
            vj = np.imag(eigvec[i]-eigvec[j])
            vi = vi/la.norm(vi)
            vj = vj/la.norm(vj)
            ei = la.norm(eigval[i])
            ej = -ei
            eigval[i],eigval[j] = ei,ej
            eigvec[i],eigvec[j] = vi,vj
        if matrix_rank(np.real(eigvec)) != len(eigvec):
            print "rank conversion failed!", matrix_rank(np.real(eigvec)), len(eigvec) ### need to test more for domains
            
    sortedargs = np.argsort(np.real(eigval))
    return np.real(eigval)[sortedargs], np.real(eigvec)[sortedargs]

def modes_sparse_symm(mx,smallest=_default_smallest,which='LM',remove_sp=True,**kwargs):
    if remove_sp and smallest % 2 == 0: smallest += 1  # remove_sp returns only pairs, gets rid of unpaired state
    eigval,eigvec = las.eigsh(mx,smallest,sigma=_close_to_zero,which=which,**kwargs)
    eigvec = np.array(eigvec).T

    if remove_sp:
        def distmeas(i,j): return np.fabs((i+j)/np.min([np.fabs(i),np.fabs(j)]))
        
        upairs = []
        ens = eigval[:]
        while(len(upairs)<len(ens)/2):
            allpairs = np.array([(i,j,distmeas(ens[i],ens[j])) for i in range(len(ens)) for j in range(i) if (i not in np.ravel(upairs) and j not in np.ravel(upairs))])
            i,j,dummy = np.array(allpairs[np.argmin(allpairs[:,2])],dtype=int)
            upairs.append((i,j))
            
        indices = np.unique(np.ravel(upairs))
        if len(indices) != smallest-1:
            print "spurious mx removal failed!", indices, eigval,[(i,np.min(np.fabs((eigval+en)/en))) for i,en in enumerate(eigval)]
        eigval = eigval[indices]
        eigvec = eigvec[indices]
    
    return eigval, eigvec

def dynamicalmodes(mesh,smallest=None):
    return modes(dynamicalmatrix(mesh))

def tensionmodes(mesh,smallest=None):
    return modes(tensionmatrix(mesh))

def ipr(eigenstate):
    '''inverse participation ratio as measure of localization'''
    eigenstate = eigenstate/np.sqrt(np.sum(eigenstate**2))
    return np.sum(eigenstate**4)
    
def ipr_zm(eigenstate):
    return ipr(eigenstate[:len(eigenstate)/2])
    
def ipr_ss(eigenstate):
    return ipr(eigenstate[len(eigenstate)/2:])

def sort_by_ipr(eigenstuff,n=None):
    if n is None: n = len(eigenstuff[0])
    iprs = np.array(all_iprs(eigenstuff[1]))
    sortedargs = np.argsort(iprs)[::-1][:n]
    return eigenstuff[0][sortedargs], eigenstuff[1][sortedargs], iprs[sortedargs]

def all_iprs(estates):
    return [ipr(estate) for estate in estates]

class RigidityObject(object):
    def __init__(self,mesh,sparse=False,normed=False,pinpoints=None):
        self.mesh = mesh
        self.sparse = sparse
        self.normed = normed
        self.rm = None
        self.dmodes = None
        self.tmodes = None
        self.hmodes = None
        self.pinpoints = pinpoints
        if self.pinpoints is not None:
            # warning: modifies point indices in the mesh
            lastidx = self.mesh.N-1
            swapspots = [i for i in range(self.mesh.N-len(pinpoints),self.mesh.N) if i not in pinpoints]
            swapindices = [i for i in pinpoints if i < (self.mesh.N-len(pinpoints))]
            for i,pt in enumerate(swapindices):
                self.mesh.swap_points(pt, swapspots[i])
    
    def rigiditymatrix(self):
        if self.rm is not None: return self.rm
        if self.sparse:
            self.rm = rigiditymatrix_sparse(self.mesh,self.normed) if self.pinpoints is None else rigiditymatrix_sparse(self.mesh,self.normed)[:,:-2*len(self.pinpoints)]
            return self.rm
        else:
            self.rm = rigiditymatrix(self.mesh,self.normed) if self.pinpoints is None else rigiditymatrix(self.mesh,self.normed)[:,:-2*len(self.pinpoints)]
            return self.rm
            
    def dynamicalmatrix(self):  # Yujie, sandwich spring matrix
        r = self.rigiditymatrix()
        return r.T*self.springmatrix()*r                # matrix multiplication

    # def dynamicalmatrix(self):
    #     r = self.rigiditymatrix()
    #     return r.T*r                # matrix multiplication
            
    def tensionmatrix(self):
        r = self.rigiditymatrix()
        if self.pinpoints is not None:
            # need to get rid of bonds with full zero entries in rigidity matrix
            rsq = r.multiply(r) if self.sparse else np.multiply(r,r)
            nonzeroed_bonds = np.where(rsq.dot(np.ones(r.shape[1])) != 0)[0]
            r = r[nonzeroed_bonds]
            print  "nz",len(nonzeroed_bonds)
            
            # visualize
            msnip = deepcopy(self.mesh)
            msnip.Bonds = [self.mesh.Bonds[i] for i in nonzeroed_bonds]
            vis2d(msnip)
            
            
        return r*r.T                # matrix multiplication

    def massmatrix(self):       # Yujie
        return np.matrix(np.identity(2*self.mesh.N))

    def dampingmatrix(self,ratio=0.01): # Yujie
        return ratio*np.matrix(np.identity(2*self.mesh.N))

    def springmatrix(self):
        K = []
        bds = self.mesh.Bonds
        for i in range(len(bds)):
            row = np.zeros(len(bds))
            row[i]=self.mesh.Bonds[i].k
            K.append(row)
        return np.matrix(K)

    def hamiltonianmatrix(self):    # kane-lubensky doubled hamiltonian, my r is their q^T
        r = self.rigiditymatrix()
        rt = r.T
        if self.sparse:
            z = csc_matrix(r.shape,dtype=r.dtype)
            return csc_matrix(bmat([[z,rt],[r,z]]))
        else:
            z = np.asmatrix(np.zeros_like(r))
            return np.asmatrix(np.bmat('z,rt;r,z'))              
            
    def dynamicalmodes(self,smallest=None):
        if self.sparse:
            if smallest is None: smallest = _default_smallest
            self.dmodes = modes_sparse(self.dynamicalmatrix(),smallest)
        else:
            self.dmodes = modes(self.dynamicalmatrix(),smallest)
            
        return self.dmodes if self.pinpoints is None else (self.dmodes[0], zeropad(self.dmodes[1],(0,2*len(self.pinpoints))))

    def dynamicalmodes_symm(self,smallest=None,remove_sp=True):
        if self.sparse:
            if smallest is None: smallest = _default_smallest
            self.dmodes = modes_sparse_symm(self.dynamicalmatrix(),smallest,remove_sp=remove_sp)
        else:
            self.dmodes = modes_symm(self.dynamicalmatrix(),smallest)
            
        return self.dmodes if self.pinpoints is None else (self.dmodes[0], zeropad(self.dmodes[1],(0,2*len(self.pinpoints))))
    
    def tensionmodes(self,smallest=None,doubled=False):
        tmx = self.tensionmatrix()
                    
        if self.sparse:
            if smallest is None: smallest = _default_smallest
            self.tmodes = modes_sparse(tmx,smallest)
        else:
            self.tmodes = modes(tmx,smallest)
            
        if doubled:
            dtmodes = np.array([np.concatenate((np.zeros(self.mesh.N*2),tmode)) for tmode in self.tmodes[1]])
            self.tmodes = self.tmodes[0],dtmodes
            
        return self.tmodes
            
            
    def hamiltonianmodes(self,smallest=None,symm=False,**kwargs):
        '''Note: do not use symm with sparse matrices, seems to give spurious results'''
        if self.sparse:
            if smallest is None: smallest = _default_smallest
            if symm: self.hmodes = modes_sparse_symm(self.hamiltonianmatrix(),smallest,**kwargs)
            else: self.hmodes = modes_sparse(self.hamiltonianmatrix(),smallest,**kwargs)
            return self.hmodes
        else:
            if symm: self.hmodes = modes_symm(self.hamiltonianmatrix(),smallest)
            else: self.hmodes = modes(self.hamiltonianmatrix(),smallest)
            return self.hmodes
            
    def hsqmodes(self,smallest=None,symm=False,**kwargs):
        '''Note: do not use symm with sparse matrices, seems to give spurious results'''
        if self.sparse:
            if smallest is None: smallest = _default_smallest
            if symm: self.hmodes = modes_sparse_symm(self.hamiltonianmatrix()*self.hamiltonianmatrix(),smallest,**kwargs)
            else: self.hmodes = modes_sparse(self.hamiltonianmatrix()*self.hamiltonianmatrix(),smallest,**kwargs)
            return self.hmodes
        else:
            if symm: self.hmodes = modes_symm(self.hamiltonianmatrix()*self.hamiltonianmatrix(),smallest)
            else: self.hmodes = modes(self.hamiltonianmatrix()*self.hamiltonianmatrix(),smallest)
            return self.hmodes
            

class NetworkSVD:
    """
        Objects to calculate the mechanical response of a truss using the 
        approach of Pellegrino, "Structural computations with the singular value
        decomposition of the equilibrium matrix", Int. J. Solids & Struct. 30, 3025 (1993)
        which systematically identifies the loads that cannot be supported due 
        to the existence of zero modes and the incompatible strains in the
        presence of self-stresses. 
    """
    def __init__(self,mesh,pinpts=None):
        ro = RigidityObject(mesh,normed=True,pinpoints=pinpts)

        self.R = ro.rigiditymatrix()
        self.A = self.R.T  # equilibrium matrix


        # from pellegrino's SVD paper
        self.nr,self.nc = self.A.shape
        self.r = matrix_rank(self.A)
        self.m = self.nr-self.r
        self.s = self.nc-self.r

        print "DOF: %d Rank: %d ZM: %d SS: %d" % (self.nr,self.r,self.m,self.s)
        self.u,self.v,self.w = la.svd(self.A)    # eq. 7
        
    def bondstresses(self, pointforces):
        '''
            calculate stresses in the struts  from a given set of forces on each point,
            ignoring the overlap with zero modes which would be unsupportable stresses.
            assumes no self-stresses.
        '''
        return np.sum([np.dot(self.u[:,i],pointforces)*self.w[i]/self.v[i] for i in range(self.r)],axis=0) # eq. 11
        
    def displacements(self, bondstresses):
        '''
            calculate point displacement given a set of bond stresses, assuming 
            equal stiffness of all members.
            todo: extend to a stiffness matrix assigning unique member stiffnesses
        '''
        return np.sum([np.dot(self.w[i],bondstresses)*self.u[:,i]/self.v[i] for i in range(self.r)],axis=0) # eq. 11
                
class TrussSVD:
    """
        Objects to calculate the mechanical response of a truss using the 
        approach of Pellegrino, "Structural computations with the singular value
        decomposition of the equilibrium matrix", Int. J. Solids & Struct. 30, 3025 (1993)
        which systematically identifies the loads that cannot be supported due 
        to the existence of zero modes and the incompatible strains in the
        presence of self-stresses. 
    """
    def __init__(self,mesh,a=1,i=1,pinpoints=None,torsion=True):
        self.mesh = mesh
        self.pinpoints = pinpoints
        if self.pinpoints is not None:
            # warning: modifies point indices in the mesh
            lastidx = self.mesh.N-1
            swapspots = [i for i in range(self.mesh.N-len(pinpoints),self.mesh.N) if i not in pinpoints]
            swapindices = [i for i in pinpoints if i < (self.mesh.N-len(pinpoints))]
            for i,pt in enumerate(swapindices):
                self.mesh.swap_points(pt, swapspots[i])
        self.R = trussrigiditymatrix(mesh,torsion) if self.pinpoints is None else trussrigiditymatrix(mesh)[:,:-3*len(pinpoints)]
        self.A = self.R.T  # equilibrium matrix
        self.k = trusskmatrix(mesh,a,i,torsion)
        self.F = la.inv(self.k)
        
        # from pellegrino's SVD paper
        self.nr,self.nc = self.A.shape
        self.r = matrix_rank(self.A)
        self.m = self.nr-self.r
        self.s = self.nc-self.r

        print "DOF: %d Rank: %d ZM: %d SS: %d" % (self.nr,self.r,self.m,self.s)
        self.u,self.v,self.w = la.svd(self.A)    # eq. 7
        
    def dynamicalmatrix(self):
        return self.R.T*self.k*self.R
        
    def tensionmatrix(self):
        return self.k*self.R*self.R.T
        
    def stresses(self,strains=None,displacements=None,pointforces=None):
        if displacements is not None:
            strains = self.strains(displacements=displacements)
        if strains is not None:
            return np.dot(self.k,strains)
        if pointforces is not None:
            wncr = self.w[-self.s:].T
            return self.bondstresses(pointforces)+np.dot(wncr,self.redundancy_eq(pointforces))
    
    def strains(self,stresses=None,displacements=None,pointforces=None):
        if pointforces is not None:
            stresses = self.stresses(pointforces=pointforces)
        if stresses is not None:
            return np.dot(self.F,stresses)
        if displacements is not None:
            return np.dot(self.R,displacements)
            
    def displacements(self,strains=None,stresses=None,pointforces=None):
        if stresses is not None:
            strains = self.strains(stresses=stresses)
        if pointforces is not None:
            strains = self.strains(pointforces=pointforces)
        if strains is not None:
            return np.sum([np.dot(self.w[i],strains)*self.u[:,i]/self.v[i] for i in range(self.r)],axis=0) # eq. 14, ignoring zero modes
        
    def bondstresses(self, pointforces):
        '''
            calculate stresses in the struts  from a given set of forces on each point,
            ignoring the overlap with zero modes which would be unsupportable stresses.
            assumes no self-stresses.
        '''
        res=np.sum([np.dot(self.u[:,i],pointforces)*self.w[i]/self.v[i] for i in range(self.r)],axis=0) # eq. 11
        return res
        
    def redundancy_eq(self, pointforces):
        ''' use the requirement of zero overlap with incompatible strains to 
            fix the redundancy in bond stresses for overconstrained systems.
            See Pellegrino eq 13'''
        wncrT = self.w[-self.s:][:]

        leftmx = np.dot(wncrT,self.F)
        leftmx = np.dot(leftmx,wncrT.T)

        rightmx = self.bondstresses(pointforces)
        rightmx = np.dot(self.F,rightmx)
        rightmx = -np.dot(wncrT,rightmx)

        x = la.solve(leftmx,rightmx)
        return x
        
    def generalized_stresses(self,pointforces):
        wncr = self.w[:,-self.s:]
        return self.bondstresses(pointforces)+np.dot(wncr,self.redundancy_eq(pointforces))
        
    
