import numpy as np
import numpy.linalg as la
from numpy.linalg import norm
#~ from emin import emin2d_so as emin2d
# from emin import emin2d_cpp as emin2d
import matplotlib.colors as colors
import matplotlib.cm as cmx
import matplotlib.collections as mc
import pylab
from copy import deepcopy
import cPickle as pickle
import os
from glob import glob
import networkx as nx
from random import randint
import sys
import time
import subprocess

def twisting_angle_to_parameter(twisting_angle):
    return np.tan(twisting_angle)/(2*np.sqrt(3))
def twisting_parameter_to_angle(parameter):
    return np.arctan(2*np.sqrt(3)*parameter)

class Logger(object):
    def __init__(self, dirname='./',filename="Default.log"):
        self.terminal = sys.stdout
        self.log = open(dirname+filename, "a")

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)
        
    def close(self):
        self.log.close()

class Log():
    def __init__(self, dirname='./',filename="Default.log"):
        self.logger = Logger(dirname,filename)
        self.logfile = dirname+filename
    
    def on(self):
        self.oldstdout = sys.stdout
        sys.stdout = self.logger
        print
        print
        print "##############################"
        print "# topomesh.py logger #"
        print "## run on",time.ctime(), "##"
        print "##############################"
        print
    
    def trim_end(self):
        linepos = 4
        cliplines = 7
        checkline = subprocess.check_output('tail -n %d %s | head -n 1' % (linepos,self.logfile), shell=True)
        print checkline
        while subprocess.check_output('tail -n %d %s | head -n 1' % (linepos,self.logfile), shell=True) == "# topomesh.py logger #\n":
            subprocess.call('head -n -%d %s > %s.tmp' % (cliplines,self.logfile,self.logfile), shell=True)
            subprocess.call('mv %s.tmp %s' % (self.logfile,self.logfile), shell=True)
        return 0
    
    def off(self):
        sys.stdout = self.oldstdout
        self.logger.close()
        self.trim_end()

def not_same(bond1,bond2):
    if bond1.p1 == bond2.p1 and bond1.p2 == bond2.p2: return False
    if bond1.p2 == bond2.p1 and bond1.p1 == bond2.p2: return False
    return True

def rotmx(theta):
    return np.array([[np.cos(theta),np.sin(theta)],[-np.sin(theta),np.cos(theta)]])
    
def rotate(v,theta):
    return np.dot(rotmx(theta),v)

def decorate_2x2(oldmesh,basis):
    """
    Given a square mesh instance `oldmesh` and `basis` of four point positions, 
    creates a decorated square mesh with every lattice point replaced by the basis.
    This takes care of decorating dislocations with Burgers vectors 
    :math:`\pm(\mathbf{a}_1\pm\mathbf{a}_2)` in terms of primitive vectors :math:`\mathbf{a}_i`.
    
    :param oldmesh: :obj:`Mesh` object, possibly including dislocations. No checks made.
    :param basis: list of four 2D vectors (arrays with 2 elements) containing the positions of the four points in the following order.
    
    ::
    
            |  |
          --3--2--
            |  |
          --0--1--
            |  |
            
    :returns: The decorated Mesh object
    """
    
    
    def newidx(oldidx):
        return 4*oldidx+np.arange(4)
    newmesh = Mesh()
    for i,pt in enumerate(oldmesh.Points):
        decoration = 2*(pt+basis)
        for newpt in decoration:
            newmesh.add_point(newpt)
        newidxs = newidx(i)
        newmesh.add_bond(newidxs[0],newidxs[1],'c')
        newmesh.add_bond(newidxs[3],newidxs[2],'m')
        newmesh.add_bond(newidxs[1],newidxs[2],'y')
        newmesh.add_bond(newidxs[0],newidxs[3],'k')
    
    newbds = []
    for bd in oldmesh.Bonds:
        newidxs1 = newidx(bd.p1)
        newidxs2 = newidx(bd.p2)
        if (bd.color == 'we'):
            newmesh.add_bond(newidxs1[1],newidxs2[0],'cc')
            newmesh.add_bond(newidxs1[2],newidxs2[3],'mm')
        if (bd.color == 'sn'):
            newmesh.add_bond(newidxs1[2],newidxs2[1],'yy')
            newmesh.add_bond(newidxs1[3],newidxs2[0],'kk')
        if (bd.color == 'my'):
            newmesh.add_bond(newidxs1[2],newidxs2[0],'cc')
            newmesh.add_bond(newidxs1[3],newidxs2[3],'mm')
        if (bd.color == 'ck'):
            newmesh.add_bond(newidxs1[1],newidxs2[1],'yy')
            newmesh.add_bond(newidxs1[2],newidxs2[0],'kk')
        if (bd.color == 'cm'):
            newmesh.add_bond(newidxs1[1],newidxs2[3],'cc')
            newmesh.add_bond(newidxs1[2],newidxs2[2],'mm')
        if (bd.color == 'ky'):
            newmesh.add_bond(newidxs1[0],newidxs2[0],'yy')
            newmesh.add_bond(newidxs1[1],newidxs2[3],'kk')
    
    newmesh.lx = oldmesh.lx*2
    newmesh.ly = oldmesh.ly*2
    return newmesh
    
def basis_bondlengths(basis):
    """Calculate bond rest lengths for a 2x2 decoration of the square lattice
    
    :param basis: Point positions of the four points in the unit cell, following
        the definition used by :func:`decorate_2x2`.
    :returns: Dictionary of the eight bond lengths identified by a one-character or two-character key.
    """
    
    bls = dict([])
    basis0 = 2*basis
    basisx = 2*(basis+np.array([1,0]))
    basisy = 2*(basis+np.array([0,1]))
    
    # bonds within basis
    bls['c'] = norm(basis0[0]-basis0[1])
    bls['m'] = norm(basis0[3]-basis0[2])
    bls['y'] = norm(basis0[1]-basis0[2])
    bls['k'] = norm(basis0[0]-basis0[3])
    
    # bonds with unit cell to right
    bls['cc'] = norm(basis0[1]-basisx[0])
    bls['mm'] = norm(basis0[2]-basisx[3])
    bls['yy'] = norm(basis0[2]-basisy[1])
    bls['kk'] = norm(basis0[3]-basisy[0])
    
    return bls

def assign_bond_lengths(mesh,color_to_length=None):
    """ Go through all the bonds in the input mesh, assigning rest lengths based on either
        an input dictionary or based on the point positions in the mesh.
        
        :param mesh: :obj:`Mesh` instance
        :param color_to_length: :obj:`dict` instance with the keys being the `color` member of the :obj:`Bond` object (set
            while the mesh was being created) and the value being the corresponding rest length to be assigned for bonds
            of the particular color. If :const:`None`, the rest lengths are calculated from the point
            positions in the input mesh.
    """
    if color_to_length is None: color_to_length = dict([]) # calculates bond lengths from mesh
    def getcolor(col):
        if col in color_to_length.keys(): return color_to_length[col]
        else: return -1
        
    for bd in mesh.Bonds:
        l0 = getcolor(bd.color)
        if l0 > 0: bd.l0 = l0
        else: bd.l0 = np.linalg.norm(mesh.dr(bd.p1,bd.p2))

def decorate_2x2_relax(oldmesh,basis):
    """Same as :func:`decorate_2x2`, but with a relaxation step after bondlengths have been assigned
    as calculated by :func:`basis_bondlengths(basis)`.
    """
    newmesh = decorate_2x2(oldmesh,basis)
    bls = basis_bondlengths(basis)
    assign_bond_lengths(newmesh,bls)
    newmesh.minimize()
    return newmesh

def klbasis(x1,x2,x3,z=0):
    """Calculates the *s* vectors in the Kane-Lubensky parametrization (*Nat Phys 2014*) of the
    deformed kagome lattices.
    
    :param x1,x2,x3,z: The four parameters used by Kane and Lubensky in their parametrization.
    :returns: List of the three vectors *s1,s2,s3* as defined in Fig. 2a of the paper.
    """ 
    a1,a2,a3=[np.array([np.cos(2*np.pi*p/3.),np.sin(2*np.pi*p/3.)]) for p in range(3)]
    y1 = z/3.+x3-x2
    y2 = z/3.+x1-x3
    s1 = x1*(a3-a2)+y1*a1
    s2 = x2*(a1-a3)+y2*a2
    return s2,np.array([0.,0.]),-s1  # return xvec30, xvec41, xvec52

def klbasispoints(x1,x2,x3,z=0):
    """Calculate the positions of the three points in the unit cell of the deformed kagome
    lattice under the Kane-Lubensky parametrization. 
    
    :param x1,x2,x3,z: Four parameters used in the Kane-Lubensky parametrization.
    :returns: List of three position vectors *d3,d1,d2* of the basis points in the unit cell as
        defined in Fig. 2a of the paper.
    """
    xvec30,xvec41,xvec52 = klbasis(x1,x2,x3,z)
    x1, x2, x3 = np.vstack((np.cos(np.pi/3*np.arange(3)),np.sin(np.pi/3*np.arange(3)))).T
    p1 = x1/2.+xvec30
    p2 = (x1+x2)/2.+xvec52
    p3 = x2/2.+xvec41
    return [p1,p2,p3]

def klbondlengths(x1,x2,x3,z=0):
    """Returns a bond length dictionary for the six bonds that define the deformed kagome unit cell."""
    xvec30,xvec41,xvec52 = klbasis(x1,x2,x3,z)
    x1, x2, x3 = np.vstack((np.cos(np.pi/3*np.arange(3)),np.sin(np.pi/3*np.arange(3)))).T
    p1 = x1/2.+xvec30
    p2 = (x1+x2)/2.+xvec52
    p3 = x2/2.+xvec41
    p4 = (x2+x3)/2.+xvec30
    p5 = x3/2.+xvec52
    
    def bl(pa,pb): return np.linalg.norm(pa-pb)
    
    colordict = dict([])
    colordict['a'] = bl(p1,p2)*2
    colordict['b'] = bl(p2,p3)*2
    colordict['c'] = bl(p3,p1)*2
    colordict['d'] = bl(p3,p4)*2
    colordict['e'] = bl(p4,p5)*2
    colordict['f'] = bl(p5,p3)*2
    
    return colordict

def decorate_kagome(mesh,x1=None,x2=None,x3=None,z=0,fraction=0.5,
                    xvec30=None,xvec41=None,xvec52=None,return_dict=False,
                    return_dicts=False,bdbyfacet=True,relax=False,relax_factor=None,
                    bondlengths=False):
    """Given an input mesh that has a hexagonal lattice including defects, decorate it
    with a 3-point unit cell to return a deformed kagome lattice. The decoration produces one 
    point in the kagome mesh from each *bond* in the input mesh using only local information
    (specifically, the positions of the endpoints of the bond in the input mesh). The philosophy is that 
    kagome mesh point corresponding to each bond in the input mesh is calculated assuming
    that the bond is part of a distortion-free hexagonal lattice; hence distortions in the original
    lattice (e.g. due to defects) lead to point positions in the final lattice that incorporate the
    degree of distortion in the input lattice although not in a totally transparent way.
    
    :param mesh: :obj:`HexMesh` object containing the base hexagonal mesh.
    :param x1,x2,x3,z: Values of Kane-Lubensky parametrization (see :func:`klbasis()` for more details).
    :param fraction, xvec30, xvec41, xvec52: Obsolete. Use default values.
    :param return_dict: If :const:True, return a dictionary that relates bond indices 
        in the input mesh to point indices in the final mesh.
    :param return_dicts: If :const:True, return two dictionaries -- the dictionary returned
        by setting ``return_dict`` to :const:True and a second dictionary that relates point
        pairs in the final mesh to the index of the bond connecting them (if it exists).
    :param bdbyfacet: Toggles between two ways of assigning the correct bonds to the final mesh. Default
        value works best and is adjusted if dictionaries need to be returned.
    :param relax: If :const:True, assign rest lengths through a call to :func:`assign_bond_lengths()` and
        then run a relaxation step by running :meth:`.Mesh.minimize()` on the final mesh.
    :param relax_factor: Parameter passed on to :meth:`.Mesh.minimize()` if parameter ``relax`` is :const:`True`.
    :param bondlengths: Assign rest lengths to all bonds in final mesh using a call to :func:`assign_bond_lengths()`.
    
    :returns: :obj:`Mesh` object containing the final kagome mesh, with ``lx`` and ``ly`` twice the box size values of 
        the input mesh. Also returns one or two dictionaries if ``return_dict`` or ``return_dicts`` is :const:`True`.
    """
    
    if (x1 is not None and x2 is not None and x3 is not None):
        '''calculate based on Kane and Lubensky notation; overrides xvecij input'''
        xvec30,xvec41,xvec52 = klbasis(x1,x2,x3,z)
    
    if relax or bondlengths: #need bond lengths to be defined
        bdbyfacet = False
    
    def ordered_pair(p1,p2):
        if (p1 < p2): return (p1,p2)
        else: return (p2,p1)
        
    def weightedavept(p1, p2, fraction=fraction):
        pt1 = np.array(mesh.Points[p1])
        pt2 = np.array(mesh.Points[p2])
        deltabond = pt1-pt2
        if deltabond[0] > mesh.lx/2: pt2[0]+=mesh.lx
        if deltabond[0] < -mesh.lx/2: pt1[0]+=mesh.lx
        if deltabond[1] > mesh.ly/2: pt2[1]+=mesh.ly
        if deltabond[1] < -mesh.ly/2: pt1[1]+=mesh.ly
        
        newpt = pt1*fraction+pt2*(1.-fraction)
        
        return newpt
        
    def innerpt(bond,xvec30=xvec30,xvec41=xvec41,xvec52=xvec52):
        p1 = bond.p1
        p2 = bond.p2
        newpt = []
        
        if xvec30 is None:
            if bond.color == '30': newpt = weightedavept(p1,p2)
            elif bond.color == '41': newpt = weightedavept(p2,p1)
            elif bond.color == '52': newpt = weightedavept(p1,p2)
            else: newpt = weightedavept(p1,p2,.5)
            
        else:
            if xvec41 is None: xvec41 = rotate(xvec30,2*np.pi/3)
            if xvec52 is None: xvec52 = rotate(xvec30,4*np.pi/3)
            if bond.color == '30': newpt = weightedavept(p1,p2)+xvec30
            elif bond.color == '41': newpt = weightedavept(p2,p1)+xvec41
            elif bond.color == '52': newpt = weightedavept(p1,p2)+xvec52
            else: newpt = weightedavept(p1,p2,.5)
        
        return newpt
    
    newmesh = Mesh()
    bd2idx = dict([])
    idxpair2bd = dict([])
    
    for bond in mesh.Bonds:
        newmesh.add_point((innerpt(bond)*2))  # displacement doubled
        bd2idx[ordered_pair(bond.p1,bond.p2)]=newmesh.N-1
    
    if bdbyfacet: #generates triangle kagome bond set for each triangle in the original tri-mesh.
        G = nx.Graph()
        G.add_edges_from(mesh.bonds())
        facets = list(nx.find_cliques(G)) #finds triangles quickly
        
        for p1,p2,p3 in facets:
            newmesh.add_bond(bd2idx[ordered_pair(p1,p2)],bd2idx[ordered_pair(p2,p3)])
            idxpair2bd[ordered_pair(bd2idx[ordered_pair(p1,p2)],bd2idx[ordered_pair(p2,p3)])] = len(newmesh.Bonds)-1
            newmesh.add_bond(bd2idx[ordered_pair(p2,p3)],bd2idx[ordered_pair(p3,p1)])
            idxpair2bd[ordered_pair(bd2idx[ordered_pair(p2,p3)],bd2idx[ordered_pair(p3,p1)])] = len(newmesh.Bonds)-1
            newmesh.add_bond(bd2idx[ordered_pair(p3,p1)],bd2idx[ordered_pair(p1,p2)])
            idxpair2bd[ordered_pair(bd2idx[ordered_pair(p3,p1)],bd2idx[ordered_pair(p1,p2)])] = len(newmesh.Bonds)-1
            
    else: #needed for different rest lengths.
        bcols = [('a','b','c'),('d','e','f'),('b','c','a'),('e','f','d'),('c','a','b'),('f','d','e')]
        for p1 in range(mesh.N):
            if len(mesh.nbrinfo[p1]) == 6:
                for j in range(6):
                    c2 = '%d' % j
                    c3 = '%d' % ((j+1) % 6)
                    p2 = mesh.nbrinfo[p1][c2]
                    p3 = mesh.nbrinfo[p1][c3]
                    
                    ca,cb,cc = bcols[j]
                    
                    b1,b2,b3 = ordered_pair(p1,p2),ordered_pair(p1,p3),ordered_pair(p2,p3)
                    newbs = ((ordered_pair(bd2idx[b1],bd2idx[b3]),ca), (ordered_pair(bd2idx[b3],bd2idx[b2]),cb), (ordered_pair(bd2idx[b2],bd2idx[b1]),cc))
                    for newb,newc in newbs:
                        if newb not in idxpair2bd:
                            newmesh.add_bond(newb[0],newb[1],color=newc)
                            idxpair2bd[newb] = len(newmesh.Bonds)-1
                        
    newmesh.lx = mesh.lx*2      # size doubled
    newmesh.ly = mesh.ly*2
    
    if relax or bondlengths:
        colordict = klbondlengths(x1,x2,x3,z)
        assign_bond_lengths(newmesh,colordict)
    if relax: # assign bond lengths and relax
        newmesh.minimize(relax_factor)
    
    if return_dict: return newmesh, bd2idx
    if return_dicts: return newmesh, bd2idx, idxpair2bd
    return newmesh

def decorate_kagome_half_x(mesh,xa,xb,leftedge=None,rightedge=None,bondlengths=False,offset=[0,0]):
    """Decorate an input mesh in such a way that there are two vertical domain walls in the structure
    between two different Kane-Lubensky unit cells.
    
    :param mesh: :class:`HexMesh` instance to be decorated.
    :param xa,xb: Tuples (x1,x2,x3,z) parametrizing the two unit cells for the outer and the central regions
        (if ``leftedge`` < ``rightedge``).
    :param leftedge: x-position of left domain wall. If :const:`None`, 1/4 of the x-length of the final mesh.
    :param rightedge: x-position of right domain wall. If :const:`None`, 3/4 of the x-length of the final mesh.
    :param bondlengths: If :const:True, assign bond lengths to the final mesh,
    :param offset: Displace each point in the region of the mesh decorated by the ``xb`` unit cell
        by this vector of two components.
        
    :returns: The resulting :class:`Mesh` instance.
    """
    newmesh1=decorate_kagome(mesh,*xa,bondlengths=bondlengths)
    newmesh2=decorate_kagome(mesh,*xb,bondlengths=bondlengths)
    
    if leftedge is None: leftedge = mesh.lx/2.
    if rightedge is None: rightedge = 3*mesh.lx/2.
    
    if leftedge < rightedge:
        for i,point in enumerate(newmesh1.Points):
            if point[0] > leftedge and point[0] < rightedge:
                point[0] = newmesh2.Points[i][0]+offset[0]
                point[1] = newmesh2.Points[i][1]+offset[1]
        if bondlengths:
            for i,bond in enumerate(newmesh1.Bonds):
                midpt = newmesh1.avept(bond.p1,bond.p2)
                if midpt[0] > leftedge and midpt[0] < rightedge:
                    bond.l0 = newmesh2.Bonds[i].l0
    else:
        for i,point in enumerate(newmesh1.Points):
            if point[0] > leftedge or point[0] < rightedge:
                point[0] = newmesh2.Points[i][0]+offset[0]
                point[1] = newmesh2.Points[i][1]+offset[1]
        if bondlengths:
            for i,bond in enumerate(newmesh1.Bonds):
                midpt = newmesh1.avept(bond.p1,bond.p2)
                if midpt[0] > leftedge or midpt[0] < rightedge:
                    bond.l0 = newmesh2.Bonds[i].l0
        
    return newmesh1

def decorate_kagome_half_y(mesh,xa,xb):
    """Similar to :func:`decorate_kagome_half_x` but with horizontal domain walls. Missing the full range 
    of functionality."""
    newmesh1=decorate_kagome(mesh,*xa)
    newmesh2=decorate_kagome(mesh,*xb)
    for i,point in enumerate(newmesh1.Points):
        if point[1] > mesh.ly/2. and point[1] < 3*mesh.ly/2.:
            point[0] = newmesh2.Points[i][0]
            point[1] = newmesh2.Points[i][1]
    return newmesh1

# helpers for sq lattice
cardinals = dict([(0,'e'),(1,'n'),(2,'w'),(3,'s')])
cardinals_rev = dict([(v,k) for k,v in cardinals.items()])
def card_shift(dirn,shift):
    return cardinals[(cardinals_rev[dirn]+shift) % 4]
    
def missing_nbr(mesh,pt):
    nbrs = mesh.nbrinfo[pt].keys()
    for dirn in ['n','s','e','w']:
        if dirn not in nbrs: return dirn
    return None


def decorate_diagonals(oldmesh,k=0.8,l0=np.sqrt(2.)):
    """Given a square lattice in ``oldmesh``, add diagonal (NNN) bonds with spring constant ``k``
    and rest length ``l0``."""
    def missing_nbr(pt):
        nbrs = oldmesh.nbrinfo[pt].keys()
        for dirn in ['n','s','e','w']:
            if dirn not in nbrs: return dirn
    
    
    def topright(idx):
        try:
            iprime = newmesh.nbrinfo[idx]['n']
            j1 = newmesh.nbrinfo[iprime]['e']
            iprime = newmesh.nbrinfo[idx]['e']
            j2 = newmesh.nbrinfo[iprime]['n']
            if (j1 == j2): return j1
            else: return -1
        except KeyError:
            return -1
    def topleft(idx):
        try:
            iprime = newmesh.nbrinfo[idx]['n']
            j1 = newmesh.nbrinfo[iprime]['w']
            iprime = newmesh.nbrinfo[idx]['w']
            j2 = newmesh.nbrinfo[iprime]['n']
            if (j1 == j2): return j1
            else: return -1
        except KeyError:
            return -1
    
    
    newmesh = deepcopy(oldmesh)
    for i in range(newmesh.N):
        j = topright(i)
        if (j >= 0): newmesh.add_bond(i,j,'dd',k,l0)
        j = topleft(i)
        if (j >= 0): newmesh.add_bond(i,j,'dd',k,l0)
        if (len(oldmesh.nbrinfo[i].keys()) == 3):
            try:
                mn = missing_nbr(i)
                p1 = oldmesh.nbrinfo[i][card_shift(mn,1)]
                p2 = oldmesh.nbrinfo[i][card_shift(mn,-1)]
                p3 = oldmesh.nbrinfo[p1][mn]
                p4 = oldmesh.nbrinfo[p2][mn]
                newmesh.add_bond(p1,p4,'dd',k,l0)
                newmesh.add_bond(p2,p3,'dd',k,l0)
            except:
                pass
            
    return newmesh
        
def nbrcount(mesh):
    return [len(nbr) for nbr in mesh.nbrinfo]


##########################
# helper for pinning
##########################
def edgepoints(mesh, border = 0.5, leftborder = None, rightborder = None, topborder = None, bottomborder = None):
    """Return indices of points in ``mesh`` within a distance ``border`` of the edge. Individual
    borders may also be specified for one or more of the left, right, top or bottom borders, which
    overrides the value of ``border`` for that particular edge."""
    if leftborder is None: leftborder = border
    if rightborder is None: rightborder = border
    if topborder is None: topborder = border
    if bottomborder is None: bottomborder = border
    
    pts = mesh.points()
    ledge, bedge = np.min(pts,axis=0)
    redge, tedge = np.max(pts,axis=0)
    border = .5
    indices = np.where((pts[:,0] < ledge + leftborder) | 
                   (pts[:,0] > redge - rightborder) | 
                   (pts[:,1] < bedge + bottomborder) | 
                   (pts[:,1] > tedge - topborder));
                   
    return indices[0]



def vis2d(mesh,draw_points=False,draw_circles=False,draw_circles_red=False,radius=20,radius_scale=1,scale=1,minlength=10**-10,draw_ptslabel=False,draw_cut=False,b1=None,b2=None,draw_annotate=None,
          draw_edges=True,eigenfunction=None,eigenfunctions=None,
            ecolors='r',figname=None,noshow=False,figidx=None,sizex=8,sizey=8,clear=False,cutoff=None,
            offset=np.array([0.,0.]),colormap='jet',colorbar=False,lw=1,quiverwidth=None,
            bondcolor=(.3,.3,.3,1.),linecollection=False,ptidx=None,ax=None,bondforces=None,
            pad=1,colorbarsymm=False,**kwargs):
    """Function for visualizing a mesh.
    
    :param mesh: Input mesh, a :class:`Mesh` instance.
    :param draw_points: If :const:`True`, draw all points with 'o' markers.
    :param draw_edges: If:const:`True`, draw bonds as lines of thickness ``lw``.
    :param eigenfunction: Pass an eigenmode for visualization. This is a 1D numpy array
        whose length is either ``2*mesh.N`` or ``2*mesh.N+Nb``, where ``N`` and ``Nb`` are the
        number of points and bonds in ``mesh``. If former, draws arrows corresponding to displacements 
        at each point (two consecutive numbers per point to give first ``2*mesh.N`` entries). 
        If latter, also colours bonds according to tensions which are the last ``Nb`` entries. 
    :param eigenfunctions: List of ``eigenfunction`` arrays to plot simultaneously.
    :param ecolors: Single color or list of colors with which to draw arrows for ``eigenfunction(s)`` visualization.
        Accepts any color specifications recognized by :mod:`matplotlib`.
    :param figname: String that, if passed, becomes filename in which visualization is saved. Should be a valid
        format for the :meth:`savefig()` command of :mod:`matplotlib.pyplot`.
    :param noshow: If :const:`True`, does not call :func:`pylab.show()` at end of plotting.
    :param figidx,ax: Specify a figure index (integer) or axis (:obj:`matplotlib.pyplot.Axes` object) within
        which to plot the mesh. Useful if you want to predefine a figure and axes to plot in. If :const:`None`,
        a blank figure and axes for the plot are created from scratch.
    :param clear: If :const:`True`, clear pre-existing stuff on ``figidx`` or ``ax``.
    :param cutoff: Length of the longest bonds that are actually drawn. If :const:`None`, assigned to be 
        the smaller of ``mesh.lx/2`` or ``mesh.ly/2``, so that bonds that wrap around the system due to 
        periodic boundary conditions are not drawn. Set to some very large number to make sure all bonds are drawn.
    :param offset: A shift to every point. Can be either a single vector (1D array with two entries) to add 
        to every point, or a ``Nx2`` array giving a unique shift to each of the ``N`` points.
    :param colormap: Colormap to use for eigenfunction tension visualization as bond colors. String (name) or ColorMap
        instance as recognized by :mod:`matplotlib`.
    :param colorbar: If :const:`True`, show a colorbar of the colormap values.
    :param lw: Line width of bond lines.
    :param quiverwidth: Width of quiver arrows used for eigenmode visualization.
    :param bondcolor: Color of bonds (if eigenfunction tensions are not defined). Any olor recognized by 
        :mod:`matplotlib` will work.
    :param linecollection: If :const:`True`, uses a slightly slower method to draw the edges which behaves 
        much better under zooming in an interactive session.
    :param ptidx: List of specific points to be drawn if ``draw_points`` is :const:`True`. 
        If :const:`None`, all points are drawn.
    :param bondforces: A 1D array of length ``Nb`` of scalar values used to color each bond. All parameters that 
        affect eigenfunction tension plotting also affect this.
    :param pad: Padding around mesh extremities to make sure all points are visible.
    :param colorbarsymm: if :const:`True`, color bar is symmetric abound zero.
    
    """
    
    if figidx is None:
        figidx = randint(0,99999)
        fig = pylab.figure(figidx,figsize=(sizex,sizey))
    else:
        fig = pylab.figure(figidx)
        if (clear): fig.clear()
    
    if ax is None: ax = fig.add_subplot(1, 1, 1, xticks=[], yticks=[])
    
        
    fig.figurePatch.set_edgecolor('white')
    fig.figurePatch.set_facecolor('white')
    ax.set_aspect('equal')
    ax.set_axis_off()
    
    pts = mesh.points()+offset
    lenmode = 2*len(pts)
    if bondforces is not None:
        # create an eigenfunction appropriately doubled to plot bond forces
        eigenfunction = np.zeros(lenmode+len(mesh.Bonds))
        eigenfunction[lenmode:] = bondforces
    drawss = (eigenfunction is not None and len(eigenfunction) == lenmode+len(mesh.Bonds))
    
    zorder=0
    
    if (draw_edges):
        if cutoff is None: cutoff = mesh.lx/2 if mesh.lx < mesh.ly else mesh.ly/2
        if not drawss:
            if linecollection:
                lines = [[(pts[bd.p1][0], pts[bd.p1][1]),(pts[bd.p2][0], pts[bd.p2][1])] for bd in mesh.Bonds if norm(pts[bd.p1]-pts[bd.p2])<cutoff]
                lc = mc.LineCollection(lines,linewidths=lw,colors=(bondcolor,),zorder=zorder)
                ax.add_collection(lc)
                xl0 = pylab.xlim()
                yl0 = pylab.ylim()
                ax.set_xlim([min(np.min(pts[:,0]),xl0[0])-pad,max(np.max(pts[:,0]),xl0[1])+pad])
                ax.set_ylim([min(np.min(pts[:,1]),yl0[0])-pad,max(np.max(pts[:,1]),yl0[1])+pad])
                zorder+=1
            else:
                xpairs = [[pts[bd.p1][0], pts[bd.p2][0]] for bd in mesh.Bonds if norm(pts[bd.p1]-pts[bd.p2])<cutoff]
                ypairs = [[pts[bd.p1][1], pts[bd.p2][1]] for bd in mesh.Bonds if norm(pts[bd.p1]-pts[bd.p2])<cutoff]
                xlist = []
                ylist = []
                for xends, yends in zip(xpairs,ypairs):
                    xlist.extend(xends)
                    xlist.append(None)
                    ylist.extend(yends)
                    ylist.append(None)
                ax.plot(xlist,ylist,'-',color=bondcolor,lw=lw,zorder=zorder)
                zorder += 1
        else:
            # use linecollection for individual colors
            ss0 = eigenfunction[lenmode:]
            colorvmin,colorvmax = np.min(ss0),np.max(ss0)
            if colorbarsymm:
                maxss0 = np.max(np.abs(ss0))
                colorvmin,colorvmax = -maxss0,maxss0
            cnorm = colors.Normalize(vmin=colorvmin,vmax=colorvmax)
            
            scalarmap = cmx.ScalarMappable(norm=cnorm,cmap=colormap)
            lines = [[(pts[bd.p1][0], pts[bd.p1][1]),(pts[bd.p2][0], pts[bd.p2][1])] for bd in mesh.Bonds if norm(pts[bd.p1]-pts[bd.p2])<cutoff]
            ssc = [scalarmap.to_rgba(ss0[i]) for i,bd in enumerate(mesh.Bonds) if norm(pts[bd.p1]-pts[bd.p2])<cutoff]
            lc = mc.LineCollection(lines,linewidths=lw,colors=ssc,zorder=zorder)
            ax.add_collection(lc)
            zorder+=1
            if colorbar:
                scalarmap.set_array(ss0)
                fig = pylab.gcf()
                fig.colorbar(scalarmap)
    
    if (draw_points):
        c=nbrcount(mesh)
        if ptidx is None:
            ax.scatter(pts[:,0],pts[:,1],c=c,**kwargs)
        else: ax.scatter(pts[:,0][ptidx],pts[:,1][ptidx],c=[c[idx] for idx in ptidx])

    if (draw_circles):
        c=['r','g','b','b','r','g']*(mesh.lx*mesh.ly/2)
        if ptidx is None:
            ax.scatter(pts[:,0],pts[:,1],s=radius*radius_scale,c=c,**kwargs)
        else: ax.scatter(pts[:,0][ptidx],pts[:,1][ptidx],s=radius*radius_scale,c=np.array(c)[ptidx],**kwargs)

        
    if (draw_circles_red):
        if ptidx is None:
            ax.scatter(pts[:,0],pts[:,1],s=radius*radius_scale,c='r',**kwargs)
        else: ax.scatter(pts[:,0][ptidx],pts[:,1][ptidx],s=radius*radius_scale,c='r',**kwargs)


    # if (draw_circles):
    #     c=['r','g','b','b','r','g']
    #     if ptidx is None:
    #         ax.scatter(pts[:,0],pts[:,1],s=radius*radius_scale,c=c,**kwargs)
    #     else:
    #         for i in ptidx:
    #             ax.scatter(pts[:,0][i],pts[:,1][i],s=radius*radius_scale,c=c[(i%6)],**kwargs)
        
    # if (draw_circles):
    #     c='r'
    #     if ptidx is None:
    #         ax.scatter(pts[:,0],pts[:,1],s=radius*radius_scale,c=c,**kwargs)
    #     else: ax.scatter(pts[:,0][ptidx],pts[:,1][ptidx],s=radius*radius_scale,c=c,**kwargs)

    # if (draw_ptslabel):         # Yujie, old
    #     for i in range(len(pts)):
    #         ax.annotate(i,(pts[i,0],pts[i,1]))

    if (draw_ptslabel):         # Yujie, select draw
        if ptidx is None:
            for i in range(len(pts)):
                ax.annotate(i,(pts[i,0],pts[i,1]))
        else:
            for i in ptidx:
                ax.annotate(i,(pts[i,0],pts[i,1]))


    if (draw_cut):              # Yujie, when at up and right boundary, not properly drawn, shifted to middle, as if not cut
        for i in range(len(b1)):
            if b1[i]!=0 and b2[i]!=0:
                # ax.annotate(i, (pts[b1[i],0]+0.5*mesh.dr(b2[i],b1[i])[0],pts[b1[i],1]+0.5*mesh.dr(b2[i],b1[i])[1]) )
                ax.annotate(i, (mesh.mid_pos(b1[i],b2[i])[0],mesh.mid_pos(b1[i],b2[i])[1]))

    if (draw_annotate!=None):              # Yujie
            ax.annotate(draw_annotate,(0.5,0.9),textcoords='figure fraction')

    # if (draw_test):
    #     print("draw_test")
    #     for nx in range(lx):
    #         for ny in range(ly):
    #             for bondlabel in range(1,6):
    #                 for labelprime in range(2):
    #                     test1,test2=kagome_bond_by_points_pair(nx,ny,bondlabel,labelprime,lx,ly)
    #                     ax.annotate(test1,(pts[test1,0],pts[test1,1]))
    #                     ax.annotate(test2,(pts[test2,0],pts[test2,1]))
            
    if (eigenfunction is not None):
        ax.quiver(pts[:,0], pts[:,1], eigenfunction[:lenmode:2], eigenfunction[1:lenmode:2],color=ecolors,zorder=zorder,width=quiverwidth,scale=scale,minlength=minlength,**kwargs)
        zorder += 1
    
    if (eigenfunctions is not None):
        for i,eig in enumerate(eigenfunctions):
            ax.quiver(pts[:,0], pts[:,1], eig[::2], eig[1::2],color=ecolors[i],zorder=zorder,width=quiverwidth,scale=scale,minlength=minlength,**kwargs)
        zorder +=1
    if figname is None and not noshow: pylab.show()
    if figname is not None:
        pylab.savefig(figname)
        pylab.close()
    
    return figidx

def write_output(mesh,filename,eigenfunction=None,factor=1.):
    # Write data files for libmd
    eps = 1e-7 # avoid boundaries of box
    pts = mesh.points()+np.array([eps,eps])-np.array([mesh.lx/2.,mesh.ly/2.])
    # shift everything within box
    dx = mesh.lx*np.round(pts[:,0]/mesh.lx)
    pts[:,0] = pts[:,0]-dx
    
    dy = mesh.ly*np.round(pts[:,1]/mesh.ly)
    pts[:,1] = pts[:,1]-dy
    
    np.savetxt(filename+'.pts',pts  )
    np.savetxt(filename+'.bds',mesh.bonds_kl(),fmt='%d %d %lf %lf')
    if eigenfunction is None:
        velocities = np.zeros_like(mesh.points())
    else:
        velocities = eigenfunction[:2*mesh.N].reshape((mesh.N,2))*factor
    np.savetxt(filename+'.vel',velocities)
    np.savetxt(filename+'.box',np.array([[mesh.lx,mesh.ly]]))

############################
# helper functions for handling decoration of pickled topomesh sequences
############################

def get_integer(string, startsegment, endsegment):
    startpos = string.find(startsegment)+len(startsegment)
    endpos = string.find(endsegment,startpos)
    return int(string[startpos:endpos])
    
def get_basis_hash(filename):
    return get_integer(filename,'basis_','.txt')
    
def get_l(filename):
    return get_integer(filename,'_l','_')

def get_directory(meshesname,basisname=None):
    if basisname is None:
        dirname = meshesname[:meshesname.find('.pkl')]+'/'
    else: dirname = meshesname[:meshesname.find('.pkl')]+'_%d/' % (get_basis_hash(basisname))
    if (not os.path.exists(dirname)): os.mkdir(dirname)
    return dirname
    
def unpickle(filename):
    inputfile = open(filename,'rb')
    pickled = pickle.load(inputfile)
    inputfile.close()
    return pickled

def nupickle(data,filename):
    outputfile = open(filename,'wb')
    pickle.dump(data,outputfile,protocol=pickle.HIGHEST_PROTOCOL)
    outputfile.close()

def n_pickled(filenametemplate):
    inthack = 314159265358979
    filenametemplate = str(filenametemplate % inthack)
    filenametemplate = filenametemplate.replace('%d' % inthack, '%s')
    return len(glob(filenametemplate % "*"))

def nupickle_each(data,filenametemplate,i):
    filename = filenametemplate % i
    nupickle(data,filename)

def nupickle_all(datacoll,filenametemplate):
    for i,data in enumerate(datacoll):
        nupickle_each(datacoll[i],filenametemplate,i)

def unpickle_each(filenametemplate,i):
    filename = filenametemplate % i
    return unpickle(filename)

def unpickle_all(filenametemplate):
    ndata = n_pickled(filenametemplate)
    datacoll = []
    for i in range(ndata):
        datacoll.append(unpickle_each(filenametemplate,i))
    return datacoll

def copy_scale(mesh,factor=1.):
    """Return a copy of ``mesh`` with all distances/positions scaled by ``factor``."""
    newmesh = deepcopy(mesh)
    newmesh.Points = [pt*factor for pt in mesh.Points]
    newmesh.lx = mesh.lx*factor
    newmesh.ly = mesh.ly*factor
    return newmesh

class Bond:
    """
    Data object that stores information about each bond in a :class:`Mesh` object.
    
    
    :ivar p1,p2: Indices of points connected by the bond.
    :ivar color: A string of one or two characters, used to identify bond type. This is
        used in hexagonal and square lattice to keep track of the orientation of the 
        bond in the lattice. It is also used to keep track of the orientation of
        dislocations, which are stored as :class:`Bond` objects as well.
    :ivar k: Spring constant.
    :ivar l0: Rest length.
    
    """
    def __init__(self, p1, p2, color='r',k=1.,l0=1.):
        self.p1 = p1
        self.p2 = p2
        self.color = color
        self.k = k
        self.l0 = l0
    
    def __str__(self):
        return '{%d, %d, %1.4f, %1.4f, %s}' % (self.p1, self.p2, self.k, self.l0, self.color)
        
    def __repr__(self):
        return self.__str__()

class Mesh:
    """
    Class that stores data defining a network of points and bonds. By default, assumes periodic
    boundary conditions in a rectangular box, i.e. distances are measured modulo the length and 
    height of the box.
    
    :ivar lx,ly: Periodic box size in x and y directions
    :ivar Points: List of 1D NumPy arrays of length 2 that store the point positions. Points
        are identified by their index in this list.
    :ivar Bonds: List of :class:`Bond` objects, one for each link in the mesh.
    :ivar nbrinfo: List of dictionaries of oriented neighbours of each point. Used for
        meshes based on regular (square and kagome) lattices for which bond orientations
        are stored in the ``Bond.color`` variable.
        e.g. if a bond between points 2 and 8 is of color `'kr'`, then::
                
            nbrinfo[2]['r'] = 8
            nbrinfo[8]['k'] = 2
                
    :ivar dislocations: List of :class:`Bond` objects that identify dislocations in the mesh.
        The color attribute of each dislocation encodes its orientation.
    :ivar N: Number of points
    :ivar dllookup: Convenience dictionary to keep track of dislocations as they are moved around.
    :ivar automin, autoplot: Boolean variables that, if :const:`True`, automatically update the mesh when 
        certain changes are made in an interactive session.
    
    The initialization generates a generic :class:`Mesh` object or builds a periodic square mesh.
    Typical usage::
            
            >>> mymesh = topomesh.Mesh() # Create an empty mesh with lx,ly set to zero
            >>> mymesh = topomesh.Mesh(20,30) # Create a 20x30 regular square mesh with unit length bonds in a periodic box of size 20x30        
    
    """
    
    def __init__(self,lx=0,ly=0,automin=False,autoplot=False):
        
        self.Points = []
        self.Bonds = []
        self.nbrinfo = []
        self.dislocations = []
        self.lx = lx
        self.ly = ly
        self.N = 0
        self.dllookup = dict([])
        
        self.make_sq_mesh(lx,ly)
        
        self.automin = automin
        self.autoplot = autoplot
        if self.autoplot: self.set_autoplot()
            
        
    def set_autoplot(self,figidx=None):
        if figidx is None: figidx = randint(0,99999)
        self.figidx = figidx
        pylab.figure(self.figidx,figsize=(16,16))
        self.autoplot = True
        self.refresh_plot()
        
    def refresh_plot(self):
        vis2d(self,draw_points=True,figidx=self.figidx,clear=True)
        
    def points(self):
        """Returns a 2D Numpy array of the point positions."""
        return np.array(self.Points)
        
    def bonds(self):
        """Returns a 2D Numpy array where each row is a pair of point indices connected by a bond."""
        return np.array([[bd.p1,bd.p2] for bd in self.Bonds])
    
    def bonds_kl(self):
        """Returns a 2D Numpy array where each row has four entries: the two point indices, the spring
        constant, and the rest length of each bond in the mesh."""
        return np.array([[bd.p1,bd.p2,bd.k,bd.l0] for bd in self.Bonds])
    
    def sizes(self):
        print self.N, 'points', len(self.Points), 'bonds', len(self.Bonds)
        
    def add_dislocation(self,dl,index=None):
        self.dislocations.append(dl)
        newidx = len(self.dislocations)-1
        if index is None:
            max_idx = max(self.dllookup) if len(self.dllookup) > 0 else -1
            index = max_idx+1
        if index in self.dllookup.keys(): print "warning: dislocation index already present"
        self.dllookup[index] = newidx
        
    def remove_dislocation(self,dl):
        dlidx = [i for i,d in enumerate(self.dislocations) if not not_same(dl,d)]
        if len(dlidx) < 1: 
            print "warning: dislocation not present"
            return -1
        if len(dlidx) > 1: print "warning: duplicate dislocation entry."
        dlidx = dlidx[0]
        lastidx = len(self.dislocations)-1
        
        self.dislocations[dlidx],self.dislocations[lastidx] = self.dislocations[lastidx],self.dislocations[dlidx]
        self.dislocations = self.dislocations[:-1]
        returnkey = -1
        for key,val in self.dllookup.items():
            if val == lastidx: self.dllookup[key] = dlidx
            if val == dlidx:
                del self.dllookup[key]
                returnkey = key
        return returnkey
    
    def add_point(self,point):
        """Add a point to the mesh.
        
        :param point: 1D array of length 2 containing the x and y coordinates of the point to be added.
            No checks made about whether the point is within the periodic box.
        """
        self.Points.append(point)
        self.nbrinfo.append(dict([]))
        self.N += 1
        
    def swap_points(self,p1,p2):
        """Swap two points at indices ``p1`` and ``p2``. Takes care of bond reassignments."""
        self.Points[p1],self.Points[p2] = self.Points[p2],self.Points[p1]
        
        for bond in self.Bonds+self.dislocations:
            if bond.p1 == p1: bond.p1 = p2
            elif bond.p1 == p2: bond.p1 = p1
            if bond.p2 == p1: bond.p2 = p2
            elif bond.p2 == p2: bond.p2 = p1
        
        for i in np.unique(self.nbrinfo[p1].values()+self.nbrinfo[p2].values()):
            for key in self.nbrinfo[i].keys():
                if self.nbrinfo[i][key] == p1: self.nbrinfo[i][key] = p2
                elif self.nbrinfo[i][key] == p2: self.nbrinfo[i][key] = p1
                
        self.nbrinfo[p1],self.nbrinfo[p2] = self.nbrinfo[p2],self.nbrinfo[p1]
    
    def remove_point(self,idx):
        """Remove point indexed by ``idx`` from the mesh. Also removves all bonds 
        that contained this point."""
        # remove all bonds
        nbrlist = self.nbrinfo[idx].values()
        for nbr in nbrlist: self.remove_bond(idx, nbr)
        
        # swap point for last point and get rid of last point
        lastpt = self.N-1
        self.swap_points(idx,lastpt)
        self.Points = self.Points[:-1]
        self.nbrinfo = self.nbrinfo[:-1]
        self.N -= 1
        
        # check that swap was successful; lastpt should not exist anywhere
        for bond in self.Bonds:
            if bond.p1 == lastpt or bond.p2 == lastpt:
                print "error in remove_point!", idx, bond
                
        # remove from dislocation list
        for dl in self.dislocations:
            if (dl.p1 == lastpt or dl.p2 == lastpt): return self.remove_dislocation(dl)
            
    def remove_points(self,ptlist):
        """Remove all points whose indices are in the list ``ptlist``. More efficient 
        than removing them through multiple calls to :func:`remove_point()`."""
        if len(ptlist) == 0: return None
        for bond in self.Bonds:
            if (bond.p1 in ptlist or bond.p2 in ptlist):
                self.remove_bond(bond)
        
        # swap points for those at end, taking care of points near the end as well
        lastidx = self.N-1
        swapspots = [i for i in range(self.N-len(ptlist),self.N) if i not in ptlist]
        swapindices = [i for i in ptlist if i < (self.N-len(ptlist))]
        for i,pt in enumerate(swapindices):
            self.swap_points(pt, swapspots[i])
        
        self.Points = self.Points[:-len(ptlist)]
        self.nbrinfo = self.nbrinfo[:-len(ptlist)]
        self.N -= len(ptlist)
        
        # check that swap was successful; lastpt should not exist anywhere
        for bond in self.Bonds:
            if bond.p1 >=self.N or bond.p2 >= self.N:
                print "error in remove_point!", self.N, bond
                
        # remove from dislocation list
        for dl in self.dislocations:
            if (dl.p1 >= self.N or dl.p2>= self.N): return self.remove_dislocation(dl)
        
    def update_points(self, points):
        """Replace the entire list of points in ``self.Points`` with the points in the
        argument ``points``."""
        if (len(points) == self.N): self.Points = list(points)
        
    def append_bond(self,bond):
        """Add a :class:`Bond` object to the list of bonds."""
        self.Bonds.append(bond)
        
    def add_bond(self, p1,p2=None,color='r',k=1.,l0=1.):
        """If called with just one argument, assumes that argument is a ``Bond`` instance which
        is added to the mesh. Otherwise, add a ``Bond`` between points indexed by ``p1`` and ``p2``,
        with color, spring constant ``k`` and rest length ``l0`` determined by the remaining arguments."""
        if p2 is None:
            self.append_bond(p1,k=k,l0=l0)
        else:
            self.append_bond(Bond(p1,p2,color,k=k,l0=l0))
            
        if len(color) == 2:
            self.nbrinfo[p1][color[1]] = p2
            self.nbrinfo[p2][color[0]] = p1
        else:
            self.nbrinfo[p1][color] = p2
            self.nbrinfo[p2][color] = p1

    def make_nbr_n(self, p1, p2):
        self.add_bond(p1,p2,'sn')
        
    def make_nbr_e(self, p1, p2):
        self.add_bond(p1,p2,'we')
        
    def make_sq_mesh(self,lx,ly):
        """Create a regular square lattice of ``lx`` columns and ``ly`` rows with periodic 
        boundary conditions (i.e. the edges have bonds that wrap around the periodic box). 
        Bonds are appropriately colored so that every bond knows its orientation in the lattice
        among the four possible orientations.
        
        Direction convention for bond color assignments::
        
              n
            w-|-e
              s
        
        """
        self.lx = lx
        self.ly = ly
        pts = np.array([[i,j] for i in range(lx) for j in range(ly)])
        N = len(pts)
        
        for pt in pts:
            self.add_point(pt)
        
        for i in range(lx):
            for j in range(ly):
                pt = i*ly + j
                if (j < (ly-1)): self.make_nbr_n(pt,pt+1)
                else: self.make_nbr_n(pt,pt-(ly-1))
                self.make_nbr_e(pt,(pt+ly)%self.N)
            
    

    def remove_bond(self,p1,p2=None):
        """If called with a single argument ``p1`` (which must be a :class:`Bond` instance),
        removes all bonds in mesh that connect the two points connected by ``p1``.
        Else, removes all bonds in the mesh that connect the point indices ``p1`` and ``p2``.
        """
        if p2 is None:
            rbond = p1
        else:
            rbond = Bond(p1,p2)
            
        bondToRemove = [bd for bd in self.Bonds if not not_same(bd,rbond)]
        self.Bonds = [bd for bd in self.Bonds if not_same(bd,rbond)]
        if len(bondToRemove) == 1:
            bd = bondToRemove[0]
            if len(bd.color) == 2:
                del self.nbrinfo[bd.p1][bd.color[1]]
                del self.nbrinfo[bd.p2][bd.color[0]]

                
    def set_bond_stiffness(self,p1,p2,k=1.0):
        """If called with a single argument ``p1`` (which must be a :class:`Bond` instance),
        removes all bonds in mesh that connect the two points connected by ``p1``.
        Else, removes all bonds in the mesh that connect the point indices ``p1`` and ``p2``.
        """
        rbond = Bond(p1,p2)
            
        for bd in self.Bonds:
            if not not_same(bd,rbond):
                bd.k = k

                
    def set_bond_color(self,p1,p2,color):
        bs = len(self.Bonds)
        self.remove_bond(p1,p2)
        if (len(self.Bonds) == bs-1):
            self.add_bond(p1,p2,color)
            
    def add_bond_ordered(self,p1,p2,color):
        if (color == 'ns'): self.add_bond(p2,p1,'sn')
        if (color == 'ew'): self.add_bond(p2,p1,'we')
        if (color == 'kc'): self.add_bond(p2,p1,'ck')
        if (color == 'ym'): self.add_bond(p2,p1,'my')
        if (color == 'mc'): self.add_bond(p2,p1,'cm')
        if (color == 'yk'): self.add_bond(p2,p1,'ky')
        if (color == 'sn' or color == 'we'
            or color == 'ck' or color == 'my'
            or color == 'cm' or color == 'ky'): self.add_bond(p1,p2,color)
        
    def dlpair_by_numbers(self,p1,p2,p3,p4,p5,p6,color1='r',color2='r',color3='r'):
        '''
            1-2-3         1 2-3
            | | |    =>   |\|\|
            4-5-6         4-5 6
        '''
        self.remove_bond(p1,p2)
        self.remove_bond(p5,p6)
        self.add_bond(p1,p5,color1)
        self.add_bond(p2,p6,color2)
        if (color3 == 'ns'): self.set_bond_color(p5,p2,'sn')
        if (color3 == 'ew'): self.set_bond_color(p5,p2,'we')
        if (color3 == 'we' or color3 == 'sn'): self.set_bond_color(p2,p5,color3)
        self.add_dislocation(Bond(p1,p5,color1))
        self.add_dislocation(Bond(p2,p6,color2))
        
        if (self.automin): self.minimize_diag()
        if (self.autoplot): self.refresh_plot()
        
    def dlpair_by_numbers_2(self,p1,p2,p3,p4,p5,p6,color1='r',color2='r',color3='r'): # Yujie
        '''
            1-2-3         1-2 3
            | | |    =>   |/|/|
            4-5-6         4 5-6
        '''
        self.remove_bond(p2,p3)
        self.remove_bond(p4,p5)
        self.add_bond(p5,p3,color1)
        self.add_bond(p4,p2,color2)
        if (color3 == 'ns'): self.set_bond_color(p5,p2,'sn')
        if (color3 == 'ew'): self.set_bond_color(p5,p2,'we')
        if (color3 == 'we' or color3 == 'sn'): self.set_bond_color(p2,p5,color3)
        self.add_dislocation(Bond(p5,p3,color1))
        self.add_dislocation(Bond(p4,p2,color2))
        
        if (self.automin): self.minimize_diag()
        if (self.autoplot): self.refresh_plot()

    def glide_by_numbers(self, p1,p2,p4,p5,color1='r',color2='r',color3='r'):
        self.remove_bond(p1,p2)
        self.remove_bond(p5,p4)
        self.add_bond(p1,p5)
        self.add_bond(p2,p4)
        
        
    def glide_dl_fwd(self, dl):
        '''glides in direction c->k for a ck colored dislocation, etc.
           'ck' and 'ky' dislocations must be glid forward first.
        '''
        if dl.color == 'ck':
            p1 = dl.p2
            p2 = dl.p1
            p4 = self.nbrinfo[p1]['e']
            p5 = self.nbrinfo[p4]['n']
            self.glide_by_numbers(p1,p2,p4,p5)
            self.set_bond_color(p4,p1,'sn')
            self.set_bond_color(p1,p5,'ck')
            self.set_bond_color(p2,p4,'we')
            
        if dl.color == 'my':
            p1 = dl.p2
            p2 = dl.p1
            p4 = self.nbrinfo[p1]['n']
            p5 = self.nbrinfo[p4]['e']
            self.glide_by_numbers(p1,p2,p4,p5)
            self.set_bond_color(p4,p1,'we')
            self.set_bond_color(p1,p5,'my')
            self.set_bond_color(p2,p4,'sn')
            
        if dl.color == 'cm':
            p1 = dl.p2
            p2 = dl.p1
            p4 = self.nbrinfo[p1]['e']
            p5 = self.nbrinfo[p4]['s']
            self.glide_by_numbers(p1,p2,p4,p5)
            self.set_bond_color(p1,p4,'sn')
            self.set_bond_color(p1,p5,'cm')
            self.set_bond_color(p2,p4,'we')
            
        if dl.color == 'ky':
            p1 = dl.p2
            p2 = dl.p1
            p4 = self.nbrinfo[p1]['s']
            p5 = self.nbrinfo[p4]['e']
            self.glide_by_numbers(p1,p2,p4,p5)
            self.set_bond_color(p4,p1,'we')
            self.set_bond_color(p1,p5,'ky')
            self.set_bond_color(p4,p2,'sn')
            
        idx = self.remove_dislocation(dl)
        self.add_dislocation(Bond(p1,p5,dl.color),index=idx)
        
        if (self.automin): self.minimize_diag()
        if (self.autoplot): self.refresh_plot()
            
            
    def glide_dl_rev(self, dl):
        '''glides in direction k->c for a ck colored dislocation, etc.
           'cm' and my' dislocations must glide reverse first
        '''
        if dl.color == 'ck':
            p1 = dl.p1
            p2 = dl.p2
            p4 = self.nbrinfo[p1]['s']
            p5 = self.nbrinfo[p4]['w']
            self.glide_by_numbers(p1,p2,p4,p5)
            self.set_bond_color(p1,p4,'we')
            self.set_bond_color(p5,p1,'ck')
            self.set_bond_color(p4,p2,'sn')
            
        if dl.color == 'my':
            p1 = dl.p1
            p2 = dl.p2
            p4 = self.nbrinfo[p1]['w']
            p5 = self.nbrinfo[p4]['s']
            self.glide_by_numbers(p1,p2,p4,p5)
            self.set_bond_color(p1,p4,'sn')
            self.set_bond_color(p5,p1,'my')
            self.set_bond_color(p4,p2,'we')
            
        if dl.color == 'cm':
            p1 = dl.p1
            p2 = dl.p2
            p4 = self.nbrinfo[p1]['n']
            p5 = self.nbrinfo[p4]['w']
            self.glide_by_numbers(p1,p2,p4,p5)
            self.set_bond_color(p1,p4,'we')
            self.set_bond_color(p5,p1,'cm')
            self.set_bond_color(p2,p4,'sn')
            
        if dl.color == 'ky':
            p1 = dl.p1
            p2 = dl.p2
            p4 = self.nbrinfo[p1]['w']
            p5 = self.nbrinfo[p4]['n']
            self.glide_by_numbers(p1,p2,p4,p5)
            self.set_bond_color(p4,p1,'sn')
            self.set_bond_color(p5,p1,'ky')
            self.set_bond_color(p4,p2,'we')
            
        idx = self.remove_dislocation(dl)
        self.add_dislocation(Bond(p5,p1,dl.color),index=idx)
        
        if (self.automin): self.minimize_diag()
        if (self.autoplot): self.refresh_plot()
    
    def climb_by_numbers(self, p1,p2,p3,p4,c1,c2,c3,c4):
        self.add_point(self.avept(p2,p4))
        newpt = self.N-1
        
        self.remove_bond(p1,p2)
        self.remove_bond(p3,p4)
        self.add_bond_ordered(p1,newpt,c1)
        self.add_bond_ordered(p2,newpt,c2)
        self.add_bond_ordered(p3,newpt,c3)
        self.add_bond_ordered(p4,newpt,c4)
        
        return newpt
        
    def climb_dl_fwd(self,dl):
        '''ck advances in direction that is pi/4 away from c->k direction.
           see compbook XIV A p 101'''
        newpt = None
        if dl.color == 'ck':
            p1,p2 = dl.p1,dl.p2
            p3 = self.nbrinfo[p2]['e']
            p4 = self.nbrinfo[p3]['s']
            newpt = self.climb_by_numbers(p1,p2,p3,p4,'we','ns','kc','sn')
          
        if dl.color == 'my':
            p1,p2 = dl.p1,dl.p2
            p3 = self.nbrinfo[p2]['n']
            p4 = self.nbrinfo[p3]['w']
            newpt = self.climb_by_numbers(p1,p2,p3,p4,'sn','ew','ym','we')
          
        if dl.color == 'cm':
            p1,p2 = dl.p1,dl.p2
            p3 = self.nbrinfo[p2]['e']
            p4 = self.nbrinfo[p3]['n']
            newpt = self.climb_by_numbers(p1,p2,p3,p4,'we','sn','mc','ns')
          
        if dl.color == 'ky':
            p1,p2 = dl.p1,dl.p2
            p3 = self.nbrinfo[p2]['s']
            p4 = self.nbrinfo[p3]['w']
            newpt = self.climb_by_numbers(p1,p2,p3,p4,'ns','ew','yk','we')
          
        if newpt is not None:
            idx = self.remove_dislocation(dl)
            self.add_dislocation(Bond(newpt,p3,dl.color),index=idx)
        
        if (self.automin): self.minimize_diag()
        if (self.autoplot): self.refresh_plot()
    
    
    def dlpair_ns(self, pt):
        p1 = self.nbrinfo[pt]['s']
        p2 = pt
        p3 = self.nbrinfo[pt]['n']
        p4 = self.nbrinfo[p1]['e']
        p5 = self.nbrinfo[p2]['e']
        p6 = self.nbrinfo[p3]['e']
        
        if (self.nbrinfo[p5]['s'] != p4 or self.nbrinfo[p5]['n'] != p6): return False
    
        self.dlpair_by_numbers(p1,p2,p3,p4,p5,p6,'my','ck','ns')
        return True
        
                
    def dlpair_ew(self,pt):
        p1 = self.nbrinfo[pt]['w']
        p2 = pt
        p3 = self.nbrinfo[pt]['e']
        p4 = self.nbrinfo[p1]['s']
        p5 = self.nbrinfo[p2]['s']
        p6 = self.nbrinfo[p3]['s']
        
        if (self.nbrinfo[p5]['w'] != p4 or self.nbrinfo[p5]['e'] != p6): return False
        
        self.dlpair_by_numbers(p1,p2,p3,p4,p5,p6,'cm','ky','ew')
        return True

        
    def dlpair_we(self,pt):     # Yujie
        p1 = self.nbrinfo[pt]['w']
        p2 = pt
        p3 = self.nbrinfo[pt]['e']
        p4 = self.nbrinfo[p1]['s']
        p5 = self.nbrinfo[p2]['s']
        p6 = self.nbrinfo[p3]['s']
        
        if (self.nbrinfo[p5]['w'] != p4 or self.nbrinfo[p5]['e'] != p6): return False
        
        self.dlpair_by_numbers_2(p1,p2,p3,p4,p5,p6,'my','ck','we')
        return True

    def get_dl(self, index):
        if type(index) is str:
            for dl in self.dislocations:
                if dl.color == index: return dl
        if type(index) is int:
            if index in self.dllookup.keys(): return self.dislocations[self.dllookup[index]]  

    def shake(self,amplitude=.0001):
        """Add a random displacement of amplitude at most ``amplitude`` to each point."""
        self.Points = [pt + np.random.random(2)*amplitude for pt in self.Points]

    def minimize(self, factor=None,factorx=None,factory=None,shake=None):
        """Find a local energy minimum, assuming harmonic springs.
        
        :param factor: If not :const:`None`, expands the periodic box by ``factor`` before relaxing,
            to impose a macroscopic bulk strain of ``factor-1``.
        :param factorx, factory: Override the overall factor with independent strain factors 
            for x and y directions.
        :param shake: If not :const:`None`, randomly perturb all points with maximum amplitude set by ``shake``
            before minimizing. 
        
        """
        if shake is not None: self.shake(shake)
        if factor is not None:
            e, res = emin2d(self.points(),self.bonds_kl(),self.lx*factor,self.ly*factor)
            self.update_points(res/factor)
        elif (factorx is not None or factory is not None):
            if factorx is None: factorx = 1.
            if factory is None: factory = 1.
            e, res = emin2d(self.points(),self.bonds_kl(),self.lx*factorx,self.ly*factory)
            res2 = np.zeros_like(res)
            res2[:,0] = res[:,0]/factorx
            res2[:,1] = res[:,1]/factory
            self.update_points(res2)
        else:
            e, res = emin2d(self.points(),self.bonds_kl(),self.lx,self.ly)
            self.update_points(res)
    
    def minimize_diag(self, factor=None):
        """Only makes sense for a square lattice. Add diagonal (NNN) bonds which provide a
        finite shear modulus to the regular square lattice before minimizing."""
        dummy = decorate_diagonals(self)
        e, res = emin2d(dummy.points(),dummy.bonds_kl(),dummy.lx,dummy.ly)
        self.update_points(res)
    
    def dr(self,p1,p2):
        """Calculate the closest distance under periodic boundary conditions between 
        points ``p1`` and ``p2``.
        """
        dx,dy = self.Points[p2] - self.Points[p1]
        dx = dx if abs(dx)<0.5*self.lx else dx-abs(dx+0.5*self.lx)+abs(dx-0.5*self.lx)
        dy = dy if abs(dy)<0.5*self.ly else dy-abs(dy+0.5*self.ly)+abs(dy-0.5*self.ly)
        return np.array([dx,dy])

    def dr2(self,p1,b1,b2):                  #Yujie
        """Calculate the closest distance under periodic boundary conditions between 
        points ``p1`` and the center of the bond between points ``b1`` and ``b2``.
        """
        dx,dy = self.mid_pos(b1,b2) - self.Points[p1] 
        dx = dx if abs(dx)<0.5*self.lx else dx-abs(dx+0.5*self.lx)+abs(dx-0.5*self.lx)
        dy = dy if abs(dy)<0.5*self.ly else dy-abs(dy+0.5*self.ly)+abs(dy-0.5*self.ly)
        return np.array([dx,dy])

    
    def mid_pos(self,p1,p2):                  #Yujie
        """Calculate the middle position between points ``p1`` and ``p2``.
        """
        mx,my=self.Points[p1]+0.5*self.dr(p1,p2)
        mx = mx if abs(mx-0.5*self.lx)<0.5*self.lx else mx-abs(mx+0.5*self.lx)+abs(mx-0.5*self.lx)
        my = my if abs(my-0.5*self.ly)<0.5*self.ly else my-abs(my+0.5*self.ly)+abs(my-0.5*self.ly)
        return np.array([mx,my])

    def avept(self, p1, p2, fraction=.5):
        """Calculate the average of two points indexed by ``p1`` and ``p2``, weighted
        by ``fraction``, taking into account periodic boundary conditions."""
        pt1 = np.array(self.Points[p1])
        pt2 = np.array(self.Points[p2])
        deltabond = pt1-pt2
        if deltabond[0] > self.lx/2: pt2[0]+=self.lx
        if deltabond[0] < -self.lx/2: pt1[0]+=self.lx
        if deltabond[1] > self.ly/2: pt2[1]+=self.ly
        if deltabond[1] < -self.ly/2: pt1[1]+=self.ly
        
        newpt = pt1*fraction+pt2*(1.-fraction)
        
        return newpt
    
    def bond_forces(self):
        """Return a list of bond tensions."""
        return [(np.linalg.norm(self.dr(bond.p1,bond.p2))-bond.l0)*bond.k for bond in self.Bonds]

    def bond_energies(self):
        """Return a list of harmonic spring energies on each bond."""
        return [((np.linalg.norm(self.dr(bond.p1,bond.p2))-bond.l0)**2)*bond.k/2. for bond in self.Bonds]
        
    def spring_energy(self):
        """Return the net harmonic spring energy of the mesh."""
        return sum(self.bond_energies())
        
    def strip_bonds(self,cutoff=None):
        """Get rid of bonds that wrap around the periodic box, to get a free-boundary 
        mesh out of a periodic mesh."""
        def bl(coords,bond):
            dx = coords[bond.p1]-coords[bond.p2]
            return np.sqrt(np.sum(dx**2))
        if cutoff is None:
            cutoff = np.min((self.lx,self.ly))/2.
        self.Bonds = [bond for bond in self.Bonds if bl(self.Points,bond) < cutoff]

    def strip_bonds_y(self,cutoff=None):  # Yujie
        """Get rid of bonds that wrap along the y-direction of the boundary, i.e., bonds at x=0 and x=lx"""
        def bl(coords,bond):
            dx = coords[bond.p1]-coords[bond.p2]
            return abs(dx[0])
        if cutoff is None:
            cutoff = (self.lx)/2.
        self.Bonds = [bond for bond in self.Bonds if bl(self.Points,bond) < cutoff]

        
class HexMesh(Mesh):
    """Class for creating a hexagonal lattice, by default periodic. Inherits the
    functions of :class:`Mesh` but overrides the dislocation functionality with a 
    different set of bond colors that are oriented along the hexagonal lattice. 
    Lattice alignment is such that rows are straight lines.
    
    Basic initialization::
        
        >>> mymesh = topomesh.HexMesh(lx,ly) # Creates a hexagonal lattice with lx columns and ly rows in a commensurate box.
    
    """
    def __init__(self,lx=0,ly=0,automin=False,autoplot=False):
        Mesh.__init__(self,automin=automin,autoplot=autoplot)
        self.make_tri_mesh(lx,ly)
    
    def make_tri_mesh(self,lx,ly):
        """Creates a hexagonal lattice with lx columns and ly rows, with periodic boundary
        conditions in a rectangular box of length ``lx`` and height ``ly*sqrt(3)/2``. The 
        bond colors and nbrinfo dictionaries have complete information about the six orientations
        of bonds for the regular lattice.
        
        direction convention for bond assignments::
        
                  2--1
                 / \/ \\
                3------0
                 \ /\ /
                  4--5 
          
        """
        def parity(nbr):
            if (nbr % 2 == 0): return 1
            else: return 0
        
        self.lx = lx
        self.ly = ly*np.sqrt(3.)/2.
        delx = -0.5
        dely = np.sqrt(3.)/2.
        
        pts = np.array([[parity(j)*delx+i,j*dely] for i in range(lx) for j in range(ly)])
        N = len(pts)
        
        for pt in pts:
            self.add_point(pt)

        for i in range(lx):
            for j in range(ly):
                pt = i*ly + j
                if parity(j):
                    self.add_bond(pt,(pt+1) % N,'41')
                    self.add_bond(pt,(pt+ly) % N,'30')
                    self.add_bond(pt,(pt-ly+1) % N,'52')
                else:
                    if (j < (ly-1)):
                        self.add_bond(pt,(pt+1) % N,'52')
                        self.add_bond(pt,(pt+ly+1) % N,'41')
                    else:
                        self.add_bond(pt,pt-ly+1,'52')
                        self.add_bond(pt,(pt+1) % N,'41')
                    self.add_bond(pt,(pt+ly) % N,'30')
                    
    
    def add_bond_ordered(self,p1,p2,color):
        if (color == '03'): self.add_bond(p2,p1,'30')
        if (color == '14'): self.add_bond(p2,p1,'41')
        if (color == '25'): self.add_bond(p2,p1,'52')
        if (color == '30' or color == '41' or color == '52'): self.add_bond(p1,p2,color)
        
    def set_bond_color_ordered(self,p1,p2,color):
        if (color == '03'): self.set_bond_color(p2,p1,'30')
        if (color == '14'): self.set_bond_color(p2,p1,'41')
        if (color == '25'): self.set_bond_color(p2,p1,'52')
        if (color == '30' or color == '41' or color == '52'): self.set_bond_color(p1,p2,color)
        
    
    def cprime(self,c):
        return '%d' % ((int(c)+3)%6)
    
    def dlpair_by_numbers(self,p1,p2,p3,p4,color1='r',color2='r',color3='r'):
        '''   
           p1--p2
            \  /\
             \/  \
             p3--p4
        '''
        self.remove_bond(p2,p3)

        self.set_bond_color(p3,p1,color1)
        self.set_bond_color(p2,p4,color2)
        self.add_bond_ordered(p1,p4,color3)
        self.add_dislocation(Bond(p3,p1,color1))
        self.add_dislocation(Bond(p2,p4,color2))
        
        if (self.automin): self.minimize()
        if (self.autoplot): self.refresh_plot()
        
    
    def dlpair_ns(self,pt):
        '''   
          p1--p2
           \  /\
            \/  \
            p3--p4
        '''
        p1 = self.nbrinfo[pt]['2']
        p2 = self.nbrinfo[pt]['1']
        p3 = pt
        p4 = self.nbrinfo[pt]['0']
        
        if (self.nbrinfo[p1]['0'] != p2 or self.nbrinfo[p2]['5'] != p4): return False
        self.dlpair_by_numbers(p1,p2,p3,p4,'sn','ns','25')
        
    def dlpair_ev(self,pt):
        '''   
              p1
              /\
             /  \
            p3--p2
             \  /
              \/
              p4
        '''
        p1 = self.nbrinfo[pt]['1']
        p2 = self.nbrinfo[pt]['0']
        p3 = pt
        p4 = self.nbrinfo[pt]['5']
        
        if (self.nbrinfo[p1]['5'] != p2 or self.nbrinfo[p2]['4'] != p4): return False
        self.dlpair_by_numbers(p1,p2,p3,p4,'ve','ev','14')
        
    def dlpair_dw(self,pt):
        '''   
              p3--p1
              /\  /
             /  \/
            p4--p2
        '''
        p1 = self.nbrinfo[pt]['0']
        p2 = self.nbrinfo[pt]['5']
        p3 = pt
        p4 = self.nbrinfo[pt]['4']
        
        if (self.nbrinfo[p1]['4'] != p2 or self.nbrinfo[p2]['3'] != p4): return False
        self.dlpair_by_numbers(p1,p2,p3,p4,'wd','dw','03')
        
    def glide_by_numbers(self,pf,ps,pa,pb,dlcolor,ncolor1,ncolor2):
        '''   
          pb--ps
           \  /|
            \/ |
            pa-pf    (configuration 'sn')
        '''
        self.remove_bond(ps,pa)
        self.set_bond_color(pa,pb,dlcolor)
        self.set_bond_color_ordered(pf,ps,ncolor2)
        self.add_bond_ordered(pf,pb,ncolor1)
        
        if (self.automin): self.minimize()
        if (self.autoplot): self.refresh_plot()
    
    def glide_algo(self,dl,c1,c2,c3,c4,c5):
        pf = dl.p1
        ps = dl.p2
        pa = self.nbrinfo[pf][c1]
        pb = self.nbrinfo[ps][c1]
        if (self.nbrinfo[pa][c2] != pb or self.nbrinfo[ps][c3] != pa): return False
        self.glide_by_numbers(pf,ps,pa,pb,dl.color,c4,c5)
        return pa,pb
    
    def glide_dl_fwd(self,dl):
        pf = dl.p1
        ps = dl.p2
        if dl.color == 'ns':
            pa,pb = self.glide_algo(dl,'0','5','1','25','14')
            
        if dl.color == 'sn':
            pa,pb = self.glide_algo(dl,'3','2','4','52','41')
            
        if dl.color == 'ev':
            pa,pb = self.glide_algo(dl,'5','4','0','14','03')
            
        if dl.color == 've':
            pa,pb = self.glide_algo(dl,'2','1','3','41','30')
            
        if dl.color == 'dw':
            pa,pb = self.glide_algo(dl,'4','3','5','03','52')
            
        if dl.color == 'wd':
            pa,pb = self.glide_algo(dl,'1','0','2','30','25')
            
        idx = self.remove_dislocation(dl)
        self.add_dislocation(Bond(pa,pb,dl.color),index=idx)
    
    
    def glide_dl_rev(self,dl):
        pf = dl.p1
        ps = dl.p2
        if dl.color == 'ns':
            pa,pb = self.glide_algo(dl,'3','4','2','14','25')
            
        if dl.color == 'sn':
            pa,pb = self.glide_algo(dl,'0','1','5','41','52')
            
        if dl.color == 'ev':
            pa,pb = self.glide_algo(dl,'2','3','1','03','14')
            
        if dl.color == 've':
            pa,pb = self.glide_algo(dl,'5','0','4','30','41')
            
        if dl.color == 'dw':
            pa,pb = self.glide_algo(dl,'1','2','0','52','03')
            
        if dl.color == 'wd':
            pa,pb = self.glide_algo(dl,'4','5','3','25','30')
            
        idx = self.remove_dislocation(dl)
        self.add_dislocation(Bond(pa,pb,dl.color),index=idx)
    
    
    def climb_by_numbers(self,p1,p2,p3,p4,p5,c1,c2,c3,c4,c5,dlnewcolor):
        '''
        /*   p3 
         *   /\
         *  /  \
         * p4--p2
         * \  /|
         *  \/ |  
         * p5--p1    */
        '''
        def avept(p1, p2):
            fraction = 0.5
            pt1 = np.array(self.Points[p1])
            pt2 = np.array(self.Points[p2])
            deltabond = pt1-pt2
            if deltabond[0] > self.lx/2: pt2[0]+=self.lx
            if deltabond[0] < -self.lx/2: pt1[0]+=self.lx
            if deltabond[1] > self.ly/2: pt2[1]+=self.ly
            if deltabond[1] < -self.ly/2: pt1[1]+=self.ly
            newpt = pt1*fraction+pt2*(1.-fraction)
            return newpt
         
        self.remove_bond(p2,p4)
        self.remove_bond(p2,p5)
        self.add_point(avept(p2,p4))
        newpt = self.N-1
        
        self.add_bond_ordered(newpt,p1,c1)
        self.add_bond_ordered(newpt,p2,c2)
        self.add_bond(newpt,p3,c3)             # new dislocation, so c3 should have order of dlcolor
        self.add_bond_ordered(newpt,p4,c4)
        self.add_bond_ordered(newpt,p5,c5)
        
        self.set_bond_color(p1,p2,dlnewcolor)
        
        if (self.automin): self.minimize()
        if (self.autoplot): self.refresh_plot()

        return newpt,p3
    
    
    
    def climb_algo(self,dl,c1,c2,c3):
        p1 = dl.p1
        p2 = dl.p2
        p5 = self.nbrinfo[p1][c1]
        p4 = self.nbrinfo[p2][c1]
        p3 = self.nbrinfo[p2][c2]
        if (self.nbrinfo[p5][c2] != p4 or 
                self.nbrinfo[p2][c3] != p5 or self.nbrinfo[p3][c3] != p4): return False
        ca = c2+self.cprime(c2)
        cb = c1+self.cprime(c1)
        cc = dl.color
        cd = self.cprime(c1)+c1
        ce = self.cprime(c3)+c3
        cnew = c3 + self.cprime(c3)
        return self.climb_by_numbers(p1,p2,p3,p4,p5,ca,cb,cc,cd,ce,cnew)
        
    def climb_dl_fwd(self,dl):
        pf = dl.p1
        ps = dl.p2
        if dl.color == 'ns':
            pa,pb = self.climb_algo(dl,'0','5','1')
            
        if dl.color == 'sn':
            pa,pb = self.climb_algo(dl,'3','2','4')
            
        if dl.color == 'ev':
            pa,pb = self.climb_algo(dl,'5','4','0')
            
        if dl.color == 've':
            pa,pb = self.climb_algo(dl,'2','1','3')
            
        if dl.color == 'dw':
            pa,pb = self.climb_algo(dl,'4','3','5')
            
        if dl.color == 'wd':
            pa,pb = self.climb_algo(dl,'1','0','2')
            
        idx = self.remove_dislocation(dl)
        self.add_dislocation(Bond(pa,pb,dl.color),index=idx)
        
    def climb_dl_rev(self,dl):
        pf = dl.p1
        ps = dl.p2
        if dl.color == 'ns':
            pa,pb = self.climb_algo(dl,'3','4','2')
            
        if dl.color == 'sn':
            pa,pb = self.climb_algo(dl,'0','1','5')
            
        if dl.color == 'ev':
            pa,pb = self.climb_algo(dl,'2','3','1')
            
        if dl.color == 've':
            pa,pb = self.climb_algo(dl,'5','0','4')
            
        if dl.color == 'dw':
            pa,pb = self.climb_algo(dl,'1','2','0')
            
        if dl.color == 'wd':
            pa,pb = self.climb_algo(dl,'4','5','3')
            
        idx = self.remove_dislocation(dl)
        self.add_dislocation(Bond(pa,pb,dl.color),index=idx)
        
    def unclimb_by_numbers(self,pf,p1,p2,p3,p4,p5,c1,c2,dlcolor):
        '''
        /*   p3 
         *   /\
         *  /  \
         * p4--p2
         * \  /|
         *  \/ |  
         * p5--p1    */
            
            does not remove the point pf!
        
        '''
        for p in (p1,p2,p3,p4,p5): self.remove_bond(pf,p)
        
        self.set_bond_color(p1,p2,dlcolor)
        self.add_bond_ordered(p2,p4,c1)
        self.add_bond_ordered(p2,p5,c2)
        
        if (self.automin): self.minimize()
        if (self.autoplot): self.refresh_plot()

        return p1,p2
    
    
    def unclimb_algo(self,dl,c1,c2,c3):
        pf = dl.p1
        ps = dl.p2
        p1 = self.nbrinfo[pf][c1]
        p2 = self.nbrinfo[pf][c2]
        p3 = ps
        p4 = self.nbrinfo[pf][self.cprime(c2)]
        p5 = self.nbrinfo[pf][c3]
        if (self.nbrinfo[p2][c3] != p1 or 
                self.nbrinfo[p3][c1] != p2 or self.nbrinfo[p3][c3] != p4 or
                self.nbrinfo[p4][c1] != p5 or self.nbrinfo[p5][c2] != p1): return False
        ca = c2+self.cprime(c2)
        cb = self.cprime(c3)+c3
        pa,pb = self.unclimb_by_numbers(pf,p1,p2,p3,p4,p5,ca,cb,dl.color)
        
        # remove point, and correct for possible index changes
        idx = self.remove_point(pf)  # captures index of destroyed dislocation
        if pa == self.N: pa = pf
        if pb == self.N: pb = pf
        self.add_dislocation(Bond(pa,pb,dl.color),idx)
        
    def unclimb_dl_fwd(self,dl): 
        if dl.color == 'ns': self.unclimb_algo(dl,'2','3','1')
        if dl.color == 'sn': self.unclimb_algo(dl,'5','0','4')
        if dl.color == 'ev': self.unclimb_algo(dl,'1','2','0')
        if dl.color == 've': self.unclimb_algo(dl,'4','5','3')
        if dl.color == 'dw': self.unclimb_algo(dl,'0','1','5')
        if dl.color == 'wd': self.unclimb_algo(dl,'3','4','2')
            
        
    def unclimb_dl_rev(self,dl):
        if dl.color == 'ns': self.unclimb_algo(dl,'1','0','2')
        if dl.color == 'sn': self.unclimb_algo(dl,'4','3','5')
        if dl.color == 'ev': self.unclimb_algo(dl,'0','5','1')
        if dl.color == 've': self.unclimb_algo(dl,'3','2','4')
        if dl.color == 'dw': self.unclimb_algo(dl,'5','4','0')
        if dl.color == 'wd': self.unclimb_algo(dl,'2','1','3')
            

def kagome_bond_by_points_pair(nx,ny,bondlabel,labelprime,lx,ly): # yujie
    """
    Return a bond's points pair 
    The bond is determined by:
         1. its unit-cell-global-position (nx,ny) as in orthogonal coordinates,
            left bottom as origin (0,0)
         2. bond's incell local position (bondlabel,labelprime)

    Parameter range:
    nx: 0 to lx-1 
    ny: 0 to ly/2-1     as each cell has double layers along y direction, to compromise with Jayson's convention
    bondlabel: 1 to 6
    labelprime: 0 (no prime) or 1 (with prime)

    Structural motif: 
    p1-p6 : 6 points
    b1-b6': 12 bonds
                   .
                  / \
              b5'/   \b6'
       p5 __b3'_/_b4'_\
          \    /p6
        b1'\  /b2'
          p4\/ 
            /\
         b5/  \b6
  p1__b3__/_b4_\
    \    /p3
   b1\  /b2
      \/
      p2

    With this choice, vis2d() fully shows first unit cell at left bottom without extra tweak.

Jayson's convention:
                       .
                      / \
                     /   \
      p3__ p4 ______/_____\
              \    /
               \  /
              p5\/ 
                /\
               /  \
p2___ p0______/____\
        \    /
         \  /
          \/
          p1
    
    """
    # get my p1 - p6 in values of Jayson's convention
    p1=3*nx*ly+6*ny 
    p2=p1+1
    p3=p1+3*ly+2 # p3 and p6 belong to the next cell in Jayson's definition, so need to shift
    p4=p1+5
    p5=p1+4
    p6=p1+3*ly+3


    if labelprime==0:
        options = {
            1:(p1,p2),
            2:(p2,p3),
            3:(p1,p3),
            4:(p3,p1+3*ly),
            5:(p3,p4),
            6:(p1+3*ly,p4),
            }
    if labelprime==1:
        if ny!=ly/2-1:
            options = {
                1:(p4,p5),
                2:(p4,p6),
                3:(p5,p6),
                4:(p6,p5+3*ly),
                5:(p6,p2+3*ly+6),
                6:(p5+3*ly,p2+3*ly+6),
                }
        if ny==ly/2-1:
            options = {
                1:(p4,p5),
                2:(p4,p6),
                3:(p5,p6),
                4:(p6,p5+3*ly),
                5:(p6,p2+6),
                6:(p5+3*ly,p2+6),
                }            
        
    return options[bondlabel][0]% (3*lx*ly),options[bondlabel][1] % (3*lx*ly)

def kagome_vertical_mode_generator(kagomemesh,ly,v_return_direction,nx_row):
        v_return=np.zeros((1,2*kagomemesh.N)) # why did I define it in as a 1*N matrix form instead of a vector form? This leads to the usage of v_return[0] instead of v_return
        for i in range(3*ly):
            if v_return_direction=="x":
                v_return[0][2*i+6*nx_row*ly]=1
            if v_return_direction=="y":
                v_return[0][2*i+6*nx_row*ly+1]=1                
        return v_return/la.norm(v_return)

def kagome_vertical_mode_generator_half_up_half_down(kagomemesh,ly,v_return_direction,nx_row):
        v_return=np.zeros((1,2*kagomemesh.N))
        for i in range(3*ly):
            if i<3*ly/2:
                if v_return_direction=="x":
                    v_return[0][2*i+6*nx_row*ly]=-1
                if v_return_direction=="y":
                    v_return[0][2*i+6*nx_row*ly+1]=-1
            else:
                if v_return_direction=="x":
                    v_return[0][2*i+6*nx_row*ly]=1
                if v_return_direction=="y":
                    v_return[0][2*i+6*nx_row*ly+1]=1                
        return v_return/la.norm(v_return)

def kagome_vertical_mode_generator_alternating(kagomemesh,ly,v_return_direction,nx_row):
        v_return=np.zeros((1,2*kagomemesh.N))
        for i in range(3*ly):
            if int(i / 6.) % 2 ==1:
                if v_return_direction=="x":
                    v_return[0][2*i+6*nx_row*ly]=-1
                if v_return_direction=="y":
                    v_return[0][2*i+6*nx_row*ly+1]=-1
            else:
                if v_return_direction=="x":
                    v_return[0][2*i+6*nx_row*ly]=1
                if v_return_direction=="y":
                    v_return[0][2*i+6*nx_row*ly+1]=1                
        return v_return/la.norm(v_return)

def kagome_vertical_mode_generator_huhd_new(kagomemesh,ly,v_return_direction,nx_row):
    v_return=np.zeros((1,2*kagomemesh.N))
    for i in range(3*ly):
        if i%6==0 or i%6==3: # select the 0 and 3
            if i<3*ly/2:
                if v_return_direction=="x":
                    v_return[0][2*i+6*(nx_row)*ly]=-1
                if v_return_direction=="y":
                    v_return[0][2*i+6*(nx_row)*ly+1]=-1
            else:
                if v_return_direction=="x":
                    v_return[0][2*i+6*(nx_row)*ly]=1
                if v_return_direction=="y":
                    v_return[0][2*i+6*(nx_row)*ly+1]=1                
    return v_return/la.norm(v_return)
